import traceback

from kivy.app import App
from kivy.factory import Factory
from kivy.properties import ListProperty, BooleanProperty
from kivy.core.window import Window
from kivy.logger import Logger
from kivy.graphics import Color, Rectangle, Triangle
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import Label
from kivy.uix.image import AsyncImage
from kivy.uix.widget import Widget
from kivy.uix.button import ButtonBehavior
from kivyblocks.widgetExt import ScrollWidget
from kivyblocks.utils import CSize
from appPublic.dictObject import DictObject
from appPublic.jsonConfig import getConfig
from appPublic.registerfunction import getRegisterFunctionByName
from .baseWidget import PressableLabel, Text, HBox, VBox
from .color_definitions import getColors
from .widget_css import WidgetCSS
from .utils import alert,absurl, SUPER
from .toggleitems import PressableBox
from .threadcall import HttpClient

class EmptyBox(Label):
	def __init__(self,size_cnt=1):
		self.size_cnt = size_cnt
		siz=CSize(self.size_cnt)
		super().__init__(text='  ',font_size=siz,
				text_size=(siz,siz),
				halign='center',
				size_hint=[None,None],size=[siz,siz])

	def onSize(self,o=None,v=None):
		return

class NodeTrigger(ButtonBehavior, EmptyBox):
	open_status = BooleanProperty(False)
	color = ListProperty([1,0,0,1])
	def __init__(self, **kw):
		self.open_points = None
		self.close_points = None
		SUPER(NodeTrigger, self, kw)
		self.countPoints()
		self.bind(size=self.onSize,pos=self.onSize)

	def countPoints(self):
		self.open_points = [0,self.height, 
				self.width,self.height,
				self.width/2,0]
		self.close_points = [0,self.height, 
				0,0, self.width,

				self.height/2]
	def pointsShift(self,points):
		x = [ p + self.pos[0] if i % 2 == 0 else p + self.pos[1] \
				for i,p in enumerate(points) ]
		return x

	def onSize(self,o=None,v=None):
		self.countPoints()
		self.showOpenStatus()
	
	def on_press(self, *args):
		self.open_status = False if self.open_status else True

	def on_open_status(self, *largs):
		self.showOpenStatus()

	def on_color(self, *largs):
		self.showOpenStatus()

	def showOpenStatus(self):
		if self.close_points is None:
			return
		points = self.close_points
		if self.open_status:
			points = self.open_points
		self.canvas.clear()
		points = self.pointsShift(points)
		with self.canvas:
			Color(*self.color)
			Triangle(points=points)

class TreeNode(BoxLayout):
	def __init__(self,data,tree=None,
						parentNode=None
						):
		"""
		base TreeNode
		data:{
			if children miss, it is a leaf node,if children is a empty array, is mean need load at it is first expanded.
			}
		"""
		super(TreeNode, self).__init__(orientation='vertical',size_hint_y=None)
		self.treeObj = tree
		self.parentNode = parentNode
		self.node_level = 0
		if self.parentNode:
			self.node_level = self.parentNode.node_level + 1
		self.data = data
		self.content = None
		self.children_open = False
		self.nodes = []
		self.initChildren = False
		self.node_box = HBox(spacing=CSize(0.5),
							csscls=self.treeObj.normal_css,
							size_hint_y=None)
		self.node_box1 = BoxLayout(orientation='horizontal')
		n = data.get('children')
		self.hasChildren_nodes = True if not n is None else False
		self.children_nodes = n
		self.children_loaded = False
		self.hasChildren_nodes = data.get('children')
		if self.hasChildren_nodes:
			self.trigger = NodeTrigger(color=self.node_box.fgcolor)
			self.trigger.bind(on_press=self.toggleChildren)
			self.buildChildrenContainer()
		else:
			self.trigger = EmptyBox()
		self.node_box.add_widget(self.trigger)
		self.add_widget(self.node_box)
		self.addContent()
		self.setSize()

	def sibling(self):
		if self.parentNode is None:
			return [ i for i in self.treeObj.nodes if i != self ]
		return [ i for i in self.parentNode.nodes if i != self ]

	def selected(self):
		self.content.selected()

	def unselected(self):
		self.content.unselected()

	def setSize(self):
		if self.children_open:
			self.height = self.node_box.height + self.node_box1.height
		else:
			self.height = self.node_box.height
		self._trigger_layout()
		if self.parentNode:
			self.parentNode._trigger_layout()
		else:
			self.treeObj._inner._trigger_layout()
		
	def buildChildrenContainer(self):
		self.node_box1.add_widget(EmptyBox())
		self.children_box = BoxLayout(orientation='vertical')
		self.node_box1.add_widget(self.children_box)
		
	def setMinWidth(self,width):
		if self.node_box.width < width:
			self.node_box.width = width
		if self.node_box1.width < width:
			self.node_box1.width = width
		for n in self.nodes:
			n.setMinWidth(width)

	def buildContent(self):
		pass

	def addContent(self):
		self.buildContent()
		self.node_box.add_widget(self.content)
		self.node_box.height = self.treeObj.rowheight
		self.node_box.width = self.trigger.width + \
						self.content.width

	def buildChildren(self):
		if self.initChildren == True:
			return

		if self.data.children is None:
			return

		if self.data.children == []:
			self.treeObj.getUrlData(self.addChildren,self.data)
			return
		if len(self.nodes) == 0:
			self.addChildren(self.data.children)
		else:
			Logger.info('Tree : not need add subnodes')
		self.initChildren = True

	def childrenSizeChange(self,tn,v):
		h = 0
		w = 0
		for n in self.nodes:
			h += n.height
			w = max(w,n.width)

		self.children_box.height = h
		self.children_box.width = w
		self.node_box1.height = self.children_box.height
		self.node_box1.width = self.trigger.width + self.children_box.width
		self.setSize()

	def addChildren(self,children):
		self.data.children = children
		self.children_box.height = 0
		self.children_box.width = 0
		for c in children:
			options = {}
			options['tree'] = self.treeObj
			options['parentNode'] = self
			options['data'] = c
			tn = self.treeObj.NodeKlass(**options)
			tn.bind(size=self.childrenSizeChange)
			self.nodes.append(tn)
			self.children_box.add_widget(tn)
			self.children_box.height += tn.height
			self.children_box.width = max(self.children_box.width,tn.width)
			self.node_box1.height = self.children_box.height
			self.node_box1.width = self.trigger.width + self.children_box.width

	def addNodeByData(self,data):
		options = {}
		options['tree'] = self.treeObj
		options['parentNode'] = self
		if isinstance(data,dict):
			options['data'] = DictObject(**data)
		else:
			options['data'] = data
		w = self.treeObj.NodeKlass(**options)
		self.nodes.append(w)
		self.children_box.add_widget(w)
		self.childrenSizeChange(None,None)
		
	def addNode(self,node):
		node.parentNode = self
		self.nodes.append(node)
		self.children_box.add_widget(node)
		self.childrenSizeChange(None,None)

	def deleteNode(self,node):
		self.children_box.remove_widget(node)
		self.nodes = [ i for i in self.nodes if i != node ]
		self.childrenSizeChange(None,None)

	def delete(self):
		parentNode = self.parentNode if self.parentNode else self.treeObj
		parentNode.deleteNode(self)

	def expand(self):
		if self.treeObj.single_expand:
			if self.parentNode:
				self.parentNode.collapseall()
			else:
				self.treeObj.collapseall()
		if self.children_open:
			return
		self.toggleChildren(None)

	def collapse(self):
		if not self.children_open:
			return
		self.toggleChildren(None)

	def expandall(self):
		if not self.children_open:
			self.toggleChildren(None)
		for n in self.nodes:
			n.expandall()

	def collapseall(self):
		if self.children_open:
			self.toggleChildren(None)
		for n in self.nodes:
			n.collapseall(None)

	def moveNode(self,node,newParent=None):
		old_parent = node.parent
		if old_parent == None:
			old_parent = node.treeObj
		old_parent.deleteNode(node)
		if newParent == None:
			newParent = node.treeObj
		newParent.addNode(node)

	def toggleChildren(self,o):
		# self.treeObj.unselect_row()
		self.children_open = True if not self.children_open else False
		if self.children_open:
			self.buildChildren()
			self.add_widget(self.node_box1)
		else:
			self.remove_widget(self.node_box1)
		self.setSize()
		# when a widget remove from its parent, the get_parent_window()
		# will return a None
		# w1 = self.children_box.get_parent_window()
		# Logger.info('Tree :get_parent_window() return=%s',str(type(w1)))
	
"""
tree options
{
	"url":
	"params",
	"normal_css",
	"selected_css",
	"single_expand",
	"select_leaf_only",
	"bgcolor",
	"checkbox",
	"multplecheck",
	"idField",
	"textFiled",
	"data" # array of {children:{},...}
}
"""
class Tree(WidgetCSS, ScrollWidget):
	data = ListProperty([])
	def __init__(self,
			url=None,
			params={},
			single_expand=False,
			select_leaf_only=True,
			bgcolor=[0.2,0.2,0.2,1],
			normal_css="default",
			row_height=2,
			selected_css="selected",
			checkbox=False,
			multiplecheck=False,
			idField='id',
			textField='text',
			data=None,
			**options):
		self.url = url
		self.params = params
		self.data = data
		self.single_expand=single_expand
		self.select_leaf_only=select_leaf_only
		self.row_height=row_height
		self.rowheight = CSize(self.row_height)
		self.bgcolor = bgcolor
		self.normal_css = normal_css
		self.selected_css = selected_css
		self.checkbox = checkbox
		self.multiplecheck = multiplecheck
		self.idField = idField
		self.textField = textField
		print('options=',options)
		super(Tree, self).__init__(**options)
		self.options = DictObject(**options)
		self.nodes = []
		self.initflag = False
		self.selected_node = None
		self.buildTree()
		self.bind(size=self.onSize,pos=self.onSize)
		self.register_event_type('on_press')

	def on_press(self,*larg):
		pass

	def select_row(self, node):
		if node.hasChildren_nodes and self.select_leaf_only:
			node.toggleChildren(node)
			node.trigger.on_press()
			if node.children_open and self.single_expand:
				for n in node.sibling():
					if n.hasChildren_nodes and n.children_open:
						n.collapse()
			return
		self.unselect_row()
		self.selected_node = node
		node.selected()
		self.dispatch('on_press', node)

	def unselect_row(self):
		if self.selected_node:
			self.selected_node.unselected()
			self.selected_node = None
		
	def onSize(self,o,v=None):
		for n in self.nodes:
			n.setMinWidth(self.width)

	def setNodeKlass(self,klass):
		self.NodeKlass = klass

	def getUrlData(self,kv=None):
		hc = HttpClient()
		params = self.params.copy()
		if kv:
			for k,v in kv.items():
				if k in params.keys():	
					params[k] = v
			params['id'] = kv[self.idField]
		config = getConfig()
		app = App.get_running_app()
		url = app.absurl(self.url)
		hc.get(url,params=params,
					callback=self.dataLoaded,
					errback=self.showError)

	def showError(self,o,e):
		traceback.print_exc()
		alert(e,title='error')

	def on_data(self, *largs):
		self.addNodes()

	def buildTree(self,kv=None):
		if not hasattr(self,'NodeKlass'):
			self.NodeKlass = TreeNode

		if self.url:
			return self.getUrlData(kv)
		if len(self.data) > 0:
			self.addNodes()

	def dataLoaded(self, o, d):
		self.data = d

	def addNodes(self):
		for c in self.data:
			self.addNode(c)

	def addNode(self,data,parentNode=None):
		options = {}
		options['tree'] = self
		options['parentNode'] = None
		options['data'] = DictObject(**data)
		w = self.NodeKlass(**options)
		self.nodes.append(w)
		self.add_widget(w)
		
	def deleteNode(self,node):
		if self.selected_node == node:
			self.selected_node = None
		self.remove_widget(node)
		self.nodes = [ i for i in self.nodes if i != node ]

	def collapseall(self):
		for n in self.nodes:
			n.collapse()

	def expandall(self):
		for n in self.nodes:
			n.expand()

class TextContent(PressableLabel):
	def __init__(self,**options):
		PressableLabel.__init__(self,**options)

class TextTreeNode(TreeNode):
	def buildContent(self):
		txt = self.data.get(self.treeObj.textField,
				self.data.get(self.treeObj.idField,'defaulttext'))
		self.content = TextContent(csscls=self.treeObj.normal_css,
							text=txt,
							size_hint_y=None,
							font_size=CSize(1),
							text_size=CSize(len(txt)-1,1),
							halign='left',
							height=CSize(2))
		self.content.bind(on_press=self.onPress)
		return 
	
	def onPress(self,o,v=None):
		self.treeObj.select_row(self)

	def selected(self):
		if hasattr(self.content,'selected'):
			self.content.selected()

	def unselected(self):
		if hasattr(self.content,'unselected'):
			self.content.unselected()

class TextTree(Tree):
	def __init__(self,nodeklass=TextTreeNode,**options):
		self.NodeKlass = nodeklass
		super().__init__(**options)

class MenuTreeNode(TextTreeNode):
	def on_size(self, *args):
		self.node_box.width = self.width
		self.content.width = self.node_box.width - self.trigger.width
		self.text_widget.width = self.content.width - CSize(1)

	def selected(self):
		self.content.active(True)

	def unselected(self):
		self.content.active(False)

	def buildContent(self):
		txt = self.data.get(self.treeObj.textField,
				self.data.get(self.treeObj.idField,'defaulttext'))
		icon = self.data.get('icon')
		self.content = PressableBox(normal_css=self.treeObj.normal_css,
				actived_css=self.treeObj.selected_css,
				size_hint_y=None,
				height=self.treeObj.rowheight
		)
		self.content.orientation = 'horizontal'
		if icon:
			img = AsyncImage(source=icon,
							size_hint=(None,None),
							height=CSize(1),
							width=CSize(1))
		else:
			img = EmptyBox()
		
		self.content.add_widget(img)
		textw = Text( text=txt,
							size_hint=(None,None),
							font_size=CSize(1),
							text_size=CSize(len(txt)-1,1),
							halign='left',
							wrap=True,
							height=CSize(2),
							width=self.width)
		self.content.add_widget(textw)
		self.text_widget = textw
		self.content.bind(on_press=self.onPress)
		return 
	

"""
{
	...
	target:kkk,
	data:[
		{
			id:
			text:
			icon:
			url:
			rfname
			children
		}
	]
}
"""
class MenuTree(TextTree):
	def __init__(self, target=None, **kw):
		self.target = target
		super().__init__(nodeklass=MenuTreeNode, **kw)

	def select_row(self, node):
		super().select_row(node)
		if not node.hasChildren_nodes:
			self.menucall(node)
		
	def menucall(self, node):
		data = {}
		dw = node.data.get('datawidget')
		if dw:
			data_widget = Factory.Blocks.getWidgetById(dw)
			if data_widget:
				vn = node.data.get('datamethod', 'getValue')
				if hasattr(data_widget, vn):
					f = getattr(data_widget, vn)
					data = f()
					if not isinstance(data, dict):
						data = {}

		url = node.data.get('url')
		target = Factory.Blocks.getWidgetById(node.data.get('target',self.target),self)
		if url:
			params = node.data.get('params',{})
			params.update(data)
			blocks = Factory.Blocks()
			desc = {
				"widgettype":"urlwidget",
				"options":{
					"url":url,
					"params":params
				}
			}
			print('menucall(), params=', params)
			w = blocks.widgetBuild(desc)
			if w and target:
				target.add_widget(w)
			return 

		rfname = node.data.get('rfname')
		if rfname:
			f = getRegisterFunctionByName(rfname)
			if f:
				f(self, **data)
			return
		
		script = node.data.get('script')
		if script:
			target_name = node.data.get('target', self.target)
			target =  Factory.Blocks.getWidgetById(target_name, self)
			data.update({'self':target})
			if target:
				eval(script,data)
			return
		


class PopupMenu(BoxLayout):
	def __init__(self,target,menudesc,**opts):
		self.target = target
		self.menudesc = menudesc
		BoxLayout.__init__(self, size_hint=(0.5,0.5), **opts)
		self.menu_tree = TextTree(**menudesc)
		self.add_widget(self.menu_tree)
		self.menu_tree.bind(on_press=self.onMenuItemTouch)
		self.register_event_type('on_press')

	def on_press(self,o,v=None):
		pass

	def onMenuItemTouch(self,o,d=None,v=None):
		data = {
			'target':self.treeObj.target,
			'menudata':d
		}
		self.dispatch('on_press',data)
		self.dismiss()

	def open(self):
		Window.add_widget(self)
		self.center = Window.center

	def dismiss(self):
		Window.remove_widget(self)

	def on_touch_down(self,touch):
		if not self.collide_point(*touch.pos):
			self.dismiss()
			return False
		super().on_touch_down(touch)
		return True




import os
import sys
from kivy.resources import resource_add_path
import signal
import codecs
import json

from appPublic.jsonConfig import getConfig
from appPublic.folderUtils import ProgramPath
from appPublic.uniqueID import getID
from appPublic.rsawrap import RSA

from kivy.factory import Factory
from kivy.metrics import sp,dp,mm
from kivy.core.window import WindowBase, Window
from kivy.clock import Clock
from kivy.logger import Logger
from kivy.utils import platform
from kivy.metrics import Metrics
from kivy.app import App

import plyer

from .register import *
from .threadcall import HttpClient,Workers
from .utils import *
from .widget_css import register_css
from .version import __version__

if platform == 'android':
	from jnius import autoclass

from .android_rotation import get_rotation

Logger.info(f'KivyBlocks:version={__version__}')
def  signal_handler(signal, frame):
	app = App.get_running_app()
	if app is None:
		return
	app.workers.running = False
	app.stop()
	print('Singal handled .........')

signal.signal(signal.SIGINT, signal_handler)


class BlocksApp(App):
	def get_rotation(self):
		return get_rotation()

	def load_csses(self):
		config = getConfig()
		if not config.css:
			return

		if config.css.css_filename:
			with codecs.open(config.css.css_filename, 'r', 'utf-8') as f:
				d = json.load(f)
				self.buildCsses(d)
		if config.css.css_url:
			hc = HttpClient()
			d = hc.get(self.realurl(config.css.css_url))
			self.buildCsses(d)

	def on_rotate(self,*largs):
		self.current_rotation = Window.rotation
		Logger.info('BlocksApp:on_rotate(), largs=%s', 
						self.current_rotation)

	def buildCsses(self, dic):
		for k,v in dic.items():
			if isinstance(v,dict):
				register_css(k,v)

	def build(self):
		self.platform = platform
		self.is_desktop = platform in ['win', 'linux', 'macosx']
		config = getConfig()
		self.public_headers = {
			"platform":self.platform
		}
		# Window.borderless = True
		print('Window.dpi=', Window.dpi, 'Metrics.dpi=', Metrics.dpi)
		Window.bind(on_request_close=self.on_close)
		Window.bind(on_rotate=self.on_rotate)
		Window.bind(size=self.device_info)
		self.workers = Workers(maxworkers=config.maxworkers or 80)
		self.workers.start()
		self.load_csses()
		self.running = True
		if config.root:
			blocks = Factory.Blocks()
			x = blocks.widgetBuild(config.root)
			if x is None:
				alert('buildError,Exit', title='Error')
				return Label(text='error')
			return x
		return None

	def realurl(self, url):
		if url.startswith('https://') or url.startswith('http://'):
			return url
		config = getConfig()
		return '%s%s' % (config.uihome, url)

	def get_user_data_path(self):
		if platform == 'android':
			Environment = autoclass('android.os.Environment')
			sdpath = Environment.getExternalStorageDirectory()
			return str(sdpath)
		sdpath = App.get_running_app().user_data_dir
		return str(sdpath)

	def get_profile_name(self):
		fname = os.path.join(self.user_data_dir,'.profile.json')
		print('profile_path=', fname)
		return fname

	def default_profile(self):
		device_id = getID()
		try:
			device_id = plyer.uniqueid.id
		except:
			pass
		if isinstance(device_id,bytes):
			device_id = device_id.decode('utf-8')

		d = {
			'device_id': device_id
		}
		return d

	def write_profile(self, dic):
		fname = self.get_profile_name()
		with codecs.open(fname,'w','utf-8') as f:
			json.dump(dic,f)

	def write_default_profile(self):
		d = self.default_profile()
		self.write_profile(d)

	def read_profile(self):
		fname = self.get_profile_name()
		if not os.path.isfile(fname):
			self.write_default_profile()
		with codecs.open(fname, 'r', 'utf-8') as f:
			d = json.load(f)
			return d

	def device_info(self, *args):
		d = {
				"platform":platform,
				"width":Window.width,
				"height":Window.height
			}
		device = {
			"device_info":";".join([f'{k}={v}' for k,v in d.items()])
		}
		self.public_headers.update(device)

	def on_close(self, *args):
		self.workers.running = False
		return False




import os
import codecs
from weakref import ref
import locale
from kivy.app import App
from kivy.properties import StringProperty
from kivy.uix.label import Label
from appPublic.Singleton import SingletonDecorator
from appPublic.jsonConfig import getConfig
from .threadcall import HttpClient

@SingletonDecorator
class I18n:
	def __init__(self):
		self.kvlang={}
		self.lang = locale.getdefaultlocale()[0]
		self.loadI18n(self.lang)
		self.i18nWidgets = []

	def addI18nWidget(self,w):
		self.i18nWidgets.append(ref(w))
	
	def loadI18nFromI18nFolder(self, lang):
		config = gtConfig()
		fpath =  os.path.join(config.i18n_folder,lang,'msg.txt')
		with codecs.open(fpath,'r','utf-8') as f:
			line = f.readline()
			d = {}
			while line:
				if line.startswith('#'):
					line = readline()
					continue
				k,v = line.split(':',1)
				d.update({k:v})
				line = readline()
			return d

	def loadI18n(self,lang):
		app = App.get_running_app()
		config = getConfig()
		self.kvlang[lang] = {}
		if config.i18n_folder:
			self.kvlang[lang] = self.loadI18nFromFolder(lang)
			return

		if config.i18n_url:
			url = '%s%s/%s' % (config.uihome, config.i18n_url, lang)
			hc = HttpClient()
			d = hc.get(url)
			print('i18n() %s get data=' % url, d, type(d))
			self.kvlang[lang] = d
		
	def __call__(self,msg,lang=None):
		if lang is None:
			lang = self.lang
		d = self.kvlang.get(lang,{})
		if d is None:
			return msg
		return d.get(msg,msg)
	
	def changeLang(self,lang):
		d = self.kvlang.get(lang)
		if not d:
			self.loadI18n(lang)
		self.lang = lang
		ws = [ w for w in self.i18nWidgets if w is not None ]
		self.i18nWidgets = ws
		for w in self.i18nWidgets:
			w.changeLang(lang)
	



from kivy.clock import Clock
from kivy.app import App
from kivy.uix.boxlayout import BoxLayout
from kivy.factory import Factory
from kivy.core.window import Window
from kivy.properties import StringProperty, BooleanProperty
from kivy.uix.image import AsyncImage

from kivyblocks.utils import CSize
from kivyblocks.ready import WidgetReady

class TwoSides(WidgetReady, BoxLayout):
	panel_shape = StringProperty('landscape')
	cannt_rotation = BooleanProperty(True)
	def __init__(self,switch_image=None, 
						landscape={},
						portrait={},
						**kw):
		BoxLayout.__init__(self,**kw)
		WidgetReady.__init__(self)
		blocks = Factory.Blocks()
		self.switch_image = switch_image
		self.portrait_modal = None

		self.landscape_widget = blocks.widgetBuild(landscape)
		blocks = Factory.Blocks()
		self.portrait_widget = blocks.widgetBuild(portrait)
		self.switch_button = None
		self.switch_button_showed = False
		if self.switch_image:
			self.build_switch_image()
		self.on_size_task = None
		self.ready_task = None
		self.register_event_type('on_interactive')
		self.register_event_type('on_beforeswitch_landscape')
		self.register_event_type('on_afterswitch_landscape')
		self.register_event_type('on_beforeswitch_portrait')
		self.register_event_type('on_afterswitch_portrait')
		self.bind(fullscreen=self.show_switch_image)
		self.app = App.get_running_app()

	def set_switch_button_pos(self):
		h = self.switch_button.height
		self.switch_button.pos = (4, self.height - h - 4)

	def show_switch_image(self, o, v=None):
		def show(*args):
			if self.switch_image and self.width > self.height:
				if self.switch_button in Window.children:
					Window.remove_widget(self.switch_button)
				self.set_switch_button_pos()
				Window.add_widget(self.switch_button)
				self.switch_button_showed = True

		print('show switch_buuton ....')
		if self.fullscreen:
			Clock.schedule_once(show, 2)

	def build_switch_image(self):
		button = Factory.Blocks().widgetBuild({
			"widgettype":"PressableBox",
			"options":{
				"size_hint":(None, None),
				"size":CSize(4,4)
			},
			"subwidgets":[
				{
					"widgettype":"AsyncImage",
					"options":{
						"source":self.switch_image
					}
				}
			]
		})
		button.bind(on_press=self.switch_portrait_widget)
		self.switch_button = button

	def switch_portrait_widget(self, *args):
		def clear_modal(o, *args):
			o.clear_widgets()

		if not self.fullscreen:
			self.fullscreen = True
			return

		if not self.portrait_modal:
			y = self.height - CSize(4)
			x = self.height * y / self.width
			w = Factory.Blocks().widgetBuild({
					"widgettype":"Modal",
					"options":{
						"auto_open":False,
						"auto_dismiss":True,
						"size_hint":[None,None],
						"size":(x,y),
						"pos":((self.width - x) / 2, 0)
					}
				})
			w.bind(on_dismiss=clear_modal)
			self.portrait_modal = w
		
		if self.portrait_widget.parent:
			self.portrait_widget.parent.remove_widget(self.portrait_widget)
		self.portrait_modal.clear_widgets()
		self.portrait_modal.add_widget(self.portrait_widget)
		self.portrait_modal.open()

	def on_size(self,*args):
		if self.width >= self.height:
			if not self.landscape_widget in self.children or \
					self.panel_shape == 'portrait':
				self.dispatch('on_beforeswitch_landscape')
				self.clear_widgets()
				self.add_widget(self.landscape_widget)
				self.dispatch('on_afterswitch_landscape')
			if self.switch_button_showed and self.cannt_rotation:
				if self.switch_button in Window.children:
					Window.remove_widget(self.switch_button)
				self.set_switch_button_pos()
				Window.add_widget(self.switch_button)
			self.panel_shape = 'landscape'
		else:
			if self.portrait_widget in self.children or \
					self.panel_shape == 'landscape':
				self.dispatch('on_beforeswitch_portrait')
				self.clear_widgets()
				self.add_widget(self.portrait_widget)
				self.dispatch('on_afterswitch_portrait')
				self.dispatch('on_interactive')
			if self.switch_button in Window.children:
				Window.remove_widget(self.switch_button)
			self.cannt_rotation = False
			self.panel_shape = 'portrait'

	def on_beforeswitch_landscape(self, *args):
		pass

	def on_afterswitch_landscape(self, *args):
		pass

	def on_beforeswitch_portrait(self, *args):
		pass

	def on_afterswitch_portrait(self, *args):
		pass

	def on_interactive(self, *args):
		pass





# coding=utf-8
"""
Layer that support point clustering
===================================
"""

from math import atan, exp, floor, log, pi, sin, sqrt
from os.path import dirname, join

from kivy.lang import Builder
from kivy.metrics import dp
from kivy.properties import (
    ListProperty,
    NumericProperty,
    ObjectProperty,
    StringProperty,
)

from kivyblocks.mapview.view import MapLayer, MapMarker

Builder.load_string(
    """
<ClusterMapMarker>:
    size_hint: None, None
    source: root.source
    size: list(map(dp, self.texture_size))
    allow_stretch: True

    Label:
        color: root.text_color
        pos: root.pos
        size: root.size
        text: "{}".format(root.num_points)
        font_size: dp(18)
"""
)


# longitude/latitude to spherical mercator in [0..1] range
def lngX(lng):
    return lng / 360.0 + 0.5


def latY(lat):
    if lat == 90:
        return 0
    if lat == -90:
        return 1
    s = sin(lat * pi / 180.0)
    y = 0.5 - 0.25 * log((1 + s) / (1 - s)) / pi
    return min(1, max(0, y))


# spherical mercator to longitude/latitude
def xLng(x):
    return (x - 0.5) * 360


def yLat(y):
    y2 = (180 - y * 360) * pi / 180
    return 360 * atan(exp(y2)) / pi - 90


class KDBush:
    """
    kdbush implementation from:
    https://github.com/mourner/kdbush/blob/master/src/kdbush.js
    """

    def __init__(self, points, node_size=64):
        self.points = points
        self.node_size = node_size

        self.ids = ids = [0] * len(points)
        self.coords = coords = [0] * len(points) * 2
        for i, point in enumerate(points):
            ids[i] = i
            coords[2 * i] = point.x
            coords[2 * i + 1] = point.y

        self._sort(ids, coords, node_size, 0, len(ids) - 1, 0)

    def range(self, min_x, min_y, max_x, max_y):
        return self._range(
            self.ids, self.coords, min_x, min_y, max_x, max_y, self.node_size
        )

    def within(self, x, y, r):
        return self._within(self.ids, self.coords, x, y, r, self.node_size)

    def _sort(self, ids, coords, node_size, left, right, depth):
        if right - left <= node_size:
            return
        m = int(floor((left + right) / 2.0))
        self._select(ids, coords, m, left, right, depth % 2)
        self._sort(ids, coords, node_size, left, m - 1, depth + 1)
        self._sort(ids, coords, node_size, m + 1, right, depth + 1)

    def _select(self, ids, coords, k, left, right, inc):
        swap_item = self._swap_item
        while right > left:
            if (right - left) > 600:
                n = float(right - left + 1)
                m = k - left + 1
                z = log(n)
                s = 0.5 + exp(2 * z / 3.0)
                sd = 0.5 * sqrt(z * s * (n - s) / n) * (-1 if (m - n / 2.0) < 0 else 1)
                new_left = max(left, int(floor(k - m * s / n + sd)))
                new_right = min(right, int(floor(k + (n - m) * s / n + sd)))
                self._select(ids, coords, k, new_left, new_right, inc)

            t = coords[2 * k + inc]
            i = left
            j = right

            swap_item(ids, coords, left, k)
            if coords[2 * right + inc] > t:
                swap_item(ids, coords, left, right)

            while i < j:
                swap_item(ids, coords, i, j)
                i += 1
                j -= 1
                while coords[2 * i + inc] < t:
                    i += 1
                while coords[2 * j + inc] > t:
                    j -= 1

            if coords[2 * left + inc] == t:
                swap_item(ids, coords, left, j)
            else:
                j += 1
                swap_item(ids, coords, j, right)

            if j <= k:
                left = j + 1
            if k <= j:
                right = j - 1

    def _swap_item(self, ids, coords, i, j):
        swap = self._swap
        swap(ids, i, j)
        swap(coords, 2 * i, 2 * j)
        swap(coords, 2 * i + 1, 2 * j + 1)

    def _swap(self, arr, i, j):
        tmp = arr[i]
        arr[i] = arr[j]
        arr[j] = tmp

    def _range(self, ids, coords, min_x, min_y, max_x, max_y, node_size):
        stack = [0, len(ids) - 1, 0]
        result = []
        x = y = 0

        while stack:
            axis = stack.pop()
            right = stack.pop()
            left = stack.pop()

            if right - left <= node_size:
                for i in range(left, right + 1):
                    x = coords[2 * i]
                    y = coords[2 * i + 1]
                    if x >= min_x and x <= max_x and y >= min_y and y <= max_y:
                        result.append(ids[i])
                continue

            m = int(floor((left + right) / 2.0))

            x = coords[2 * m]
            y = coords[2 * m + 1]

            if x >= min_x and x <= max_x and y >= min_y and y <= max_y:
                result.append(ids[m])

            nextAxis = (axis + 1) % 2

            if min_x <= x if axis == 0 else min_y <= y:
                stack.append(left)
                stack.append(m - 1)
                stack.append(nextAxis)
            if max_x >= x if axis == 0 else max_y >= y:
                stack.append(m + 1)
                stack.append(right)
                stack.append(nextAxis)

        return result

    def _within(self, ids, coords, qx, qy, r, node_size):
        sq_dist = self._sq_dist
        stack = [0, len(ids) - 1, 0]
        result = []
        r2 = r * r

        while stack:
            axis = stack.pop()
            right = stack.pop()
            left = stack.pop()

            if right - left <= node_size:
                for i in range(left, right + 1):
                    if sq_dist(coords[2 * i], coords[2 * i + 1], qx, qy) <= r2:
                        result.append(ids[i])
                continue

            m = int(floor((left + right) / 2.0))

            x = coords[2 * m]
            y = coords[2 * m + 1]

            if sq_dist(x, y, qx, qy) <= r2:
                result.append(ids[m])

            nextAxis = (axis + 1) % 2

            if (qx - r <= x) if axis == 0 else (qy - r <= y):
                stack.append(left)
                stack.append(m - 1)
                stack.append(nextAxis)
            if (qx + r >= x) if axis == 0 else (qy + r >= y):
                stack.append(m + 1)
                stack.append(right)
                stack.append(nextAxis)

        return result

    def _sq_dist(self, ax, ay, bx, by):
        dx = ax - bx
        dy = ay - by
        return dx * dx + dy * dy


class Cluster:
    def __init__(self, x, y, num_points, id, props):
        self.x = x
        self.y = y
        self.num_points = num_points
        self.zoom = float("inf")
        self.id = id
        self.props = props
        self.parent_id = None
        self.widget = None

        # preprocess lon/lat
        self.lon = xLng(x)
        self.lat = yLat(y)


class Marker:
    def __init__(self, lon, lat, cls=MapMarker, options=None):
        self.lon = lon
        self.lat = lat
        self.cls = cls
        self.options = options

        # preprocess x/y from lon/lat
        self.x = lngX(lon)
        self.y = latY(lat)

        # cluster information
        self.id = None
        self.zoom = float("inf")
        self.parent_id = None
        self.widget = None

    def __repr__(self):
        return "<Marker lon={} lat={} source={}>".format(
            self.lon, self.lat, self.source
        )


class SuperCluster:
    """Port of supercluster from mapbox in pure python
    """

    def __init__(self, min_zoom=0, max_zoom=16, radius=40, extent=512, node_size=64):
        self.min_zoom = min_zoom
        self.max_zoom = max_zoom
        self.radius = radius
        self.extent = extent
        self.node_size = node_size

    def load(self, points):
        """Load an array of markers.
        Once loaded, the index is immutable.
        """
        from time import time

        self.trees = {}
        self.points = points

        for index, point in enumerate(points):
            point.id = index

        clusters = points
        for z in range(self.max_zoom, self.min_zoom - 1, -1):
            start = time()
            print("build tree", z)
            self.trees[z + 1] = KDBush(clusters, self.node_size)
            print("kdbush", (time() - start) * 1000)
            start = time()
            clusters = self._cluster(clusters, z)
            print(len(clusters))
            print("clustering", (time() - start) * 1000)
        self.trees[self.min_zoom] = KDBush(clusters, self.node_size)

    def get_clusters(self, bbox, zoom):
        """For the given bbox [westLng, southLat, eastLng, northLat], and
        integer zoom, returns an array of clusters and markers
        """
        tree = self.trees[self._limit_zoom(zoom)]
        ids = tree.range(lngX(bbox[0]), latY(bbox[3]), lngX(bbox[2]), latY(bbox[1]))
        clusters = []
        for i in range(len(ids)):
            c = tree.points[ids[i]]
            if isinstance(c, Cluster):
                clusters.append(c)
            else:
                clusters.append(self.points[c.id])
        return clusters

    def _limit_zoom(self, z):
        return max(self.min_zoom, min(self.max_zoom + 1, z))

    def _cluster(self, points, zoom):
        clusters = []
        c_append = clusters.append
        trees = self.trees
        r = self.radius / float(self.extent * pow(2, zoom))

        # loop through each point
        for i in range(len(points)):
            p = points[i]
            # if we've already visited the point at this zoom level, skip it
            if p.zoom <= zoom:
                continue
            p.zoom = zoom

            # find all nearby points
            tree = trees[zoom + 1]
            neighbor_ids = tree.within(p.x, p.y, r)

            num_points = 1
            if isinstance(p, Cluster):
                num_points = p.num_points
            wx = p.x * num_points
            wy = p.y * num_points

            props = None

            for j in range(len(neighbor_ids)):
                b = tree.points[neighbor_ids[j]]
                # filter out neighbors that are too far or already processed
                if zoom < b.zoom:
                    num_points2 = 1
                    if isinstance(b, Cluster):
                        num_points2 = b.num_points
                    # save the zoom (so it doesn't get processed twice)
                    b.zoom = zoom
                    #  accumulate coordinates for calculating weighted center
                    wx += b.x * num_points2
                    wy += b.y * num_points2
                    num_points += num_points2
                    b.parent_id = i

            if num_points == 1:
                c_append(p)
            else:
                p.parent_id = i
                c_append(
                    Cluster(wx / num_points, wy / num_points, num_points, i, props)
                )
        return clusters


class ClusterMapMarker(MapMarker):
    source = StringProperty(join(dirname(__file__), "icons", "cluster.png"))
    cluster = ObjectProperty()
    num_points = NumericProperty()
    text_color = ListProperty([0.1, 0.1, 0.1, 1])

    def on_cluster(self, instance, cluster):
        self.num_points = cluster.num_points

    def on_touch_down(self, touch):
        return False


class ClusteredMarkerLayer(MapLayer):
    cluster_cls = ObjectProperty(ClusterMapMarker)
    cluster_min_zoom = NumericProperty(0)
    cluster_max_zoom = NumericProperty(16)
    cluster_radius = NumericProperty("40dp")
    cluster_extent = NumericProperty(512)
    cluster_node_size = NumericProperty(64)

    def __init__(self, **kwargs):
        self.cluster = None
        self.cluster_markers = []
        super().__init__(**kwargs)

    def add_marker(self, lon, lat, cls=MapMarker, options=None):
        if options is None:
            options = {}
        marker = Marker(lon, lat, cls, options)
        self.cluster_markers.append(marker)
        return marker

    def remove_marker(self, marker):
        self.cluster_markers.remove(marker)

    def reposition(self):
        if self.cluster is None:
            self.build_cluster()
        margin = dp(48)
        mapview = self.parent
        set_marker_position = self.set_marker_position
        bbox = mapview.get_bbox(margin)
        bbox = (bbox[1], bbox[0], bbox[3], bbox[2])
        self.clear_widgets()
        for point in self.cluster.get_clusters(bbox, mapview.zoom):
            widget = point.widget
            if widget is None:
                widget = self.create_widget_for(point)
            set_marker_position(mapview, widget)
            self.add_widget(widget)

    def build_cluster(self):
        self.cluster = SuperCluster(
            min_zoom=self.cluster_min_zoom,
            max_zoom=self.cluster_max_zoom,
            radius=self.cluster_radius,
            extent=self.cluster_extent,
            node_size=self.cluster_node_size,
        )
        self.cluster.load(self.cluster_markers)

    def create_widget_for(self, point):
        if isinstance(point, Marker):
            point.widget = point.cls(lon=point.lon, lat=point.lat, **point.options)
        elif isinstance(point, Cluster):
            point.widget = self.cluster_cls(lon=point.lon, lat=point.lat, cluster=point)
        return point.widget

    def set_marker_position(self, mapview, marker):
        x, y = mapview.get_window_xy_from(marker.lat, marker.lon, mapview.zoom)
        marker.x = int(x - marker.width * marker.anchor_x)
        marker.y = int(y - marker.height * marker.anchor_y)



from kivy.uix.button import Button
from jnius import autoclass
from jnius import cast
from .messager import Messager

class PhoneButton(Button):
	def __init__(self,**kw):
		self.options = kw
		self.phone_number = kw.get('phone_number')
		if self.phone_number is None:
			raise Exception('PhoneButton,miss phone_number')

		del self.options['phone_number']
		self.options.update({'text':'call'})
		super(PhoneButton,self).__init__(**self.options)
		self.bind(on_release=self.makecall)

	def makecall(self,inst):
		try:
			self.disabled = True
			Intent = autoclass('android.content.Intent') 
			Uri = autoclass('android.net.Uri')
			PythonActivity = autoclass('org.kivy.android.PythonActivity')
			intent = Intent(Intent.ACTION_CALL)
			intent.setData(Uri.parse("tel:" + self.phone_number))
			currentActivity = cast('android.app.Activity',
				PythonActivity.mActivity)
			currentActivity.startActivity(intent)
		except Exception as e:
			msger = Messager()
			msger.show_error(e)



from kivy.uix.widget import Widget
from kivy.clock import Clock
from jnius import autoclass
from android.runnable import run_on_ui_thread


WebView = autoclass('android.webkit.WebView')
WebViewClient = autoclass('android.webkit.WebViewClient')
activity = autoclass('org.kivy.android.PythonActivity').mActivity


class AWebView(Widget):
	def __init__(self, url=None):
		super(AWebView, self).__init__()
		self.baseUrl = url
		self.create_webview()
		# Clock.schedule_once(self.create_webview, 0)

	@run_on_ui_thread
	def create_webview(self, *args):
		print('create_webview() begining')
		self.webview = WebView(activity)
		settings = self.webview.getSettings()
		settings.setJavaScriptEnabled(True)
		settings.setUseWideViewPort(True) # enables viewport html meta tags
		settings.setLoadWithOverviewMode(True) # uses viewport
		settings.setSupportZoom(True) # enables zoom
		settings.setBuiltInZoomControls(True) # enables zoom controls
		wvc = WebViewClient()
		self.webview.setWebViewClient(wvc)
		activity.setContentView(self.webview)
		if self.baseUrl is not None:
			self.webview.loadUrl(self.baseUrl)
			print('go to the url=',self.baseUrl)
		print('create_webview() finished')
	
	def open(self,url):
		self.webview.loadUrl(url)



from kivy.core.window import Window
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button
from kivy.uix.label import Label
from kivy.app import App
from .utils import CSize
from .bgcolorbehavior import BGColorBehavior
from .ready import WidgetReady

class PageContainer(WidgetReady, BGColorBehavior, FloatLayout):
	def __init__(self,**kw):
		FloatLayout.__init__(self, **kw)
		BGColorBehavior.__init__(self)
		WidgetReady.__init__(self)
		self.show_back = True
		self.pageWidgets = []
		self.backButton = Button(text='<',size_hint=(None,None),
							font_size=CSize(1),
							height=CSize(1.8),width=CSize(4))
		self.backButton.bind(on_press=self.previous)
		Window.bind(size=self.on_window_size)

	def on_window_size(self,o,v=None):
		if self.size[0] != Window.size[0] or self.size[1] != Window.size[1]:
			print('on_window_size event fired ....',self.size, Window.size)
			self.size = Window.size
			self.reshowBackButton()

	def reshowBackButton(self,o=None,v=None):
		print(self.size)
		self.showBackButton()

	def hideBack(self):
		if not self.show_back:
			return
		super(PageContainer,self).remove_widget(self.backButton)
		self.show_back = False

	def showBack(self):
		if self.show_back:
			return
		super(PageContainer,self).add_widget(self.backButton)
		self.backButton.text = '<%d' % len(self.pageWidgets)
		self.show_back = True

	def showLastPage(self):
		super(PageContainer,self).clear_widgets()
		if len(self.pageWidgets) < 1:
			return 
		w = self.pageWidgets[-1]
		w.pos = 0,0
		w.size = self.size
		super(PageContainer,self).add_widget(w)
		self.showBackButton()

	def previous(self,v=None):
		if len(self.pageWidgets) <= 1:
			return 
		w = self.pageWidgets[-1]
		self.pageWidgets = self.pageWidgets[:-1]
		self.showLastPage()
		if hasattr(w,'__del__'):
			w.__del__()

	def add_widget(self,widget):
		self.pageWidgets.append(widget)
		self.showLastPage()

	def showBackButton(self):
		if len(self.pageWidgets) <= 1:
			return
		print('size event fired .....',self.size)
		super(PageContainer,self).remove_widget(self.backButton)
		self.show_back = False
		self.backButton.pos = (4,self.height - 4 - self.backButton.height)
		self.showBack()


if __name__ == '__main__':
	class Page(Button):
		def __init__(self,container,page_cnt = 1):
			self.container = container
			self.page_cnt = page_cnt
			Button.__init__(self,text = 'page' + str(page_cnt))
			self.bind(on_press=self.newpage)

		def newpage(self,v):
			p = Page(self.container,self.page_cnt + 1)
			self.container.add(p)

	class MyApp(App):
		def build(self):
			x = PageContainer()
			p = Page(x)
			x.add(p)
			return x

	MyApp().run()



import os
import sys
import codecs
# import ujson as json
try:
    import ujson as json
except:
    import json
from traceback import print_exc

from functools import partial

from appPublic.dictExt import dictExtend
from appPublic.folderUtils import ProgramPath
from appPublic.dictObject import DictObject
from appPublic.Singleton import SingletonDecorator, GlobalEnv
from appPublic.datamapping import keyMapping
from appPublic.registerfunction import RegisterFunction

from kivy.logger import Logger
from kivy.config import Config
from kivy.metrics import sp,dp,mm
from kivy.core.window import WindowBase, Window
from kivy.properties import BooleanProperty
from kivy.uix.widget import Widget
from kivy.clock import mainthread
from kivy.uix.modalview import ModalView
from kivy.app import App
from kivy.factory import Factory
from kivy.uix.video import Video
from .utils import *
from .newvideo import Video
from .orientationlayout import OrientationLayout
from .threadcall import HttpClient
from .register import *
from .script import Script, set_script_env

class WidgetNotFoundById(Exception):
	def __init__(self, id):
		super().__init__()
		self.idstr = id
	
	def __str__(self):
		return "Widget not found by id:" + self.idstr + ':'

	def __expr__(self):
		return str(self)

class ClassMethodNotFound(Exception):
	def __init__(self,k,m):
		super().__init__()
		self.kname = k
		self.mname = m

	def __str__(self):
		s = 'Method(%s) not found in class(%s)' % (self.mname,
					str(self.kname.__classname__))
		return s
	
	def __expr__(self):
		return self.__str__()

	
class NotExistsObject(Exception):
	def __init__(self,name):
		super().__init__()
		self.name = name

	def __str__(self):
		s = 'not exists widget(%s)' % self.name
		return s
	
	def __expr__(self):
		return self.__str__()

class ArgumentError(Exception):
	def __init__(self,argument,desc):
		super().__init__()
		self.argument = argument
		self.desc = desc
	
	def __str__(self):
		s = 'argument(%s) missed:%s' % (self.argument,self.desc)
		return s
	
	def __expr__(self):
		return self.__str__()


class NotRegistedWidget(Exception):
	def __init__(self, name:str):
		super().__init__()
		self.widget_name = name

	def __str__(self):
		s = 'not reigsted widget(%s)' % self.name
		return s
	
	def __expr__(self):
		return self.__str__()

def registerWidget(name:str,widget):
	globals()[name] = widget


class Blocks(EventDispatcher):
	def __init__(self):
		EventDispatcher.__init__(self)
		self.action_id = 0
		self.register_event_type('on_built')
		self.register_event_type('on_failed')
		self.env = GlobalEnv()
		config = getConfig()
		self.script = Script()

	def set(self, k:str, v):
		self.env[k] = v
	
	def register_widget(self, name:str, widget:Widget):
		globals()[name] = widget

	def buildAction(self, widget:Widget, desc):
		conform_desc = desc.get('conform')
		blocks = Blocks()
		if not conform_desc:
			return partial(blocks.uniaction, widget, desc)
		func =partial(blocks.conform_action,widget, desc)
		return func
		
	def eval(self, s:str, l:dict):
		g = {}
		forbidens = [
			"os",
			"sys",
			"codecs",
			"json",
		]

		for k,v in globals().copy().items():
			if k not in forbidens:
				g[k] = v

		"""
		g['__builtins__'] = globals()['__builtins__']
		g['__builtins__']['__import__'] = None
		g['__builtins__']['__loader__'] = None
		g['__builtins__']['open'] = None
		"""
		g.update(self.env)
		return eval(s,g,l)

	def getUrlData(self, url:str, method:str='GET',
						params:dict={}, files:dict={},
					callback=None,
					errback=None,**kw):

		if url is None:
			if errback:
				errback(None,Exception('url is None'))
			else:
				return None

		if url.startswith('file://'):
			return self.script.dispatch(url, **params)
		elif url.startswith('http://') or url.startswith('https://'):
			try:
				hc = HttpClient()
				resp=hc(url,method=method,params=params,files=files)
				# print('Blocks.py :resp=',resp)
				return resp
			except Exception as e:
				print_exc()
				if errback:
					return errback(None,e)
				return None
		else:
			config = getConfig()
			url = config.uihome + url
			return self.getUrlData(url,method=method,
					params=params,
					files=files,
					**kw)

	def strValueExpr(self,s:str,localnamespace:dict={}):
		if not s.startswith('py::'):
			return s
		s = s[4:]
		try:
			v = self.eval(s,localnamespace)
			return v
		except Exception as e:
			print('Exception .... ',e,'script=',s)
			print_exc()
			return s

	def arrayValueExpr(self,arr:list,localnamespace:dict={}) -> list:
		d = []
		for v in arr:
			if type(v) == type(''):
				d.append(self.strValueExpr(v,localnamespace))
				continue
			if type(v) == type([]):
				d.append(self.arrayValueExpr(v,localnamespace))
				continue
			if type(v) == type({}):
				d.append(self.dictValueExpr(v,localnamespace))
				continue
			if type(v) == type(DictObject):
				d.append(self.dictValueExpr(v,localnamespace))
				continue
			d.append(v)
		return d

	def dictValueExpr(self,dic:dict,localnamespace:dict={}) -> dict:
		d = {}
		for k,v in dic.items():
			if type(v) == type(''):
				d[k] = self.strValueExpr(v,localnamespace)
				continue
			if type(v) == type([]):
				d[k] = self.arrayValueExpr(v,localnamespace)
				continue
			if type(v) == type({}):
				d[k] = self.dictValueExpr(v,localnamespace)
				continue
			if type(v) == type(DictObject):
				d[k] = self.dictValueExpr(v,localnamespace)
				continue
			d[k] = v
		return d
	def valueExpr(self,obj,localnamespace:dict={}):
		if type(obj) == type(''):
			return self.strValueExpr(obj,localnamespace)
		if type(obj) == type([]):
			return self.arrayValueExpr(obj,localnamespace)
		if type(obj) == type({}):
			return self.dictValueExpr(obj,localnamespace)
		if isinstance(obj,DictObject):
			return self.dictValueExpr(obj,localnamespace)
		return obj

	def w_build(self,desc) -> Widget:
		widgetClass = desc.get('widgettype',None)
		if not widgetClass:
			Logger.info("Block: w_build(), desc invalid", desc)
			raise Exception(desc)

		widgetClass = desc['widgettype']
		opts = self.valueExpr(desc.get('options',{}).copy())
		widget = None
		try:
			klass = Factory.get(widgetClass)
			widget = klass(**opts)
		except Exception as e:
			print('Error:',widgetClass,'contructon error')
			print_exc()
			raise NotExistsObject(widgetClass)

		if desc.get('id'):
			id = desc.get('id')
			if id.startswith('app.'):
				app = App.get_running_app()
				id = id[4:]
				setattr(app, id, widget)
			if id.startswith('root.'):
				app = App.get_running_app()
				id = id[5:]
				setattr(app.root, id, widget)
				
			if '.' in id:
				Logger.info('widget id(%s) can not contain "."', id)
			else:
				widget.widget_id = id
		
		widget.build_desc = desc
		self.build_attributes(widget, desc)
		self.build_rest(widget, desc)
		self.buildBinds(widget, desc)
		return widget
		
	def build_attributes(self, widget:Widget,desc,t=None):
		excludes = ['widgettype','options','subwidgets','binds']
		for k,v in [(k,v) for k,v in desc.items() if k not in excludes]:
			if isinstance(v,dict) and v.get('widgettype'):
				b = Blocks()
				v = self.valueExpr(v, localnamespace={'self':widget})
				w = b.widgetBuild(v)
				if hasattr(widget,k):
					aw = getattr(widget,k)
					if isinstance(aw,Layout):
						aw.add_widget(w)
						continue
				setattr(widget,k,w)
				continue
			setattr(widget,k,self.valueExpr(v,\
						localnamespace={'self':widget}))

	def build_rest(self, widget:Widget,desc,t=None):
		self.subwidget_total = len(desc.get('subwidgets',[]))
		self.subwidgets = [ None for i in range(self.subwidget_total)]
		pos = 0
		for pos,sw in enumerate(desc.get('subwidgets',[])):
			b = Blocks()
			if isinstance(sw, str):
				# w = Blocks.getWidgetById(sw, from_widget=widget)
				w = Blocks.findWidget(sw, from_widget=widget)
				if w:
					widget.add_widget(w)
				continue
				
			kw = self.valueExpr(sw.copy(), 
						localnamespace={'self':widget})
			w = b.widgetBuild(kw)
			if w:
				widget.add_widget(w)


	def buildBinds(self, widget:Widget, desc:dict):
		for b in desc.get('binds',[]):
			kw = self.valueExpr(b.copy(), \
						localnamespace={'self':widget})
			self.buildBind(widget,kw)

	def buildBind(self, widget:Widget, desc:dict):
		wid = desc.get('wid','self')
		# w = Blocks.getWidgetById(desc.get('wid','self'),from_widget=widget)
		w = Blocks.findWidget(desc.get('wid','self'),from_widget=widget)
		if not w:
			Logger.info('Block: id(%s) %s',desc.get('wid','self'),
							'not found via Blocks.getWidgetById()')
			return
		event = desc.get('event')
		if event is None:
			Logger.info('Block: binds desc miss event, desc=%s',str(desc))
			return
		f = self.buildAction(widget,desc)
		if f is None:
			Logger.info('Block: get a null function,%s',str(desc))
			return
		w.bind(**{event:f})
	
	def multipleAction(self, widget:Widget, desc, *args):
		desc1 = {k:v for k, v in desc.items() if k != 'actions'}
		mydesc = desc1.copy()
		for a in desc['actions']:
			new_desc = mydesc.copy()
			new_desc.update(a)
			self.uniaction(widget,new_desc, *args)

	def conform_action(self, widget:Widget, desc, *args):
		conform_desc = desc.get('conform')
		blocks = Blocks()
		if not conform_desc:
			blocks.uniaction(widget, desc,*args, **kw)
			return
		w = blocks.widgetBuild({
			"widgettype":"Conform",
			"options":conform_desc
		})
		w.bind(on_conform=partial(blocks.uniaction, widget, desc))
		w.open()

	def uniaction(self, widget:Widget, desc, *args):
		acttype = desc.get('actiontype')
		if acttype=='blocks':
			return self.blocksAction(widget,desc, *args)
		if acttype=='urlwidget':
			return self.urlwidgetAction(widget,desc, *args)
		if acttype == 'registedfunction':
			return self.registedfunctionAction(widget,desc, *args)
		if acttype == 'script':
			return self.scriptAction(widget, desc, *args)
		if acttype == 'method':
			return self.methodAction(widget, desc, *args)
		if acttype == 'event':
			return self.eventAction(widget, desc, *args)
		if acttype == 'multiple':
			return self.multipleAction(widget, desc, *args)

		alert("actiontype(%s) invalid" % acttype,title='error')

	def eventAction(self, widget:Widget, desc, *args):
		target = self.get_target(widget, desc)
		event = desc.get('dispatch_event')
		if not event:
			Logger.info('Block: eventAction():desc(%s) miss dispatch_event',
							str(desc))
			return
		params = desc.get('params',{})
		d = self.getActionData(widget,desc, *args)
		if d:
			params.update(d)
		try:
			target.dispatch(event, params)
		except Exception as e:
			Logger.info(f'Block: eventAction():dispatch {event} error')
			print_exc()
			return
		
	def get_target(self, widget:Widget, desc):
		if not desc.get('target'):
			return None
		# return Blocks.getWidgetById(desc.get('target'),from_widget=widget)
		return Blocks.findWidget(desc.get('target'),from_widget=widget)

	def blocksAction(self, widget:Widget, desc, *args):
		target = self.get_target(widget, desc)
		add_mode = desc.get('mode','replace')
		opts = desc.get('options').copy()
		d = self.getActionData(widget,desc, *args)
		p = opts.get('options',{}).copy()
		if d:
			p.update(d)
		opts['options'] = p
		def doit(target:Widget, add_mode:str, o, w:Widget):
			if isinstance(w, Modal):
				return

			if target and not w.parent:
				if add_mode == 'replace':
					target.clear_widgets()
				target.add_widget(w)

		def doerr(o,e):
			Logger.info('Block: blocksAction(): desc=%s widgetBuild error'
								,str(desc))
			raise e

		b = Blocks()
		b.bind(on_built=partial(doit,target,add_mode))
		b.bind(on_failed=doerr)
		b.widgetBuild(opts)
		
	def urlwidgetAction(self, widget:Widget, desc, *args):
		target = self.get_target(widget, desc)
		add_mode = desc.get('mode','replace')
		opts = desc.get('options', {}).copy()
		p1 = opts.get('params',{})
		p = {}
		if p1:
			p.update(p1)
		if len(args) >= 1 and isinstance(args[0],dict):
			p.update(args[0])
		d = self.getActionData(widget, desc, *args)
		if d:
			p.update(d)
		opts['params'] = p
		d = {
			'widgettype' : 'urlwidget',
			'options': opts
		}

		def doit(target:Widget, add_mode:str, o, w:Widget):
			if isinstance(w, ModalView):
				return

			if target and not w.parent:
				if add_mode == 'replace':
					target.clear_widgets()
				target.add_widget(w)

		def doerr(o,e):
			Logger.info('Block: urlwidgetAction(): desc=%s widgetBuild error'
								,str(desc))

		b = Blocks()
		b.bind(on_built=partial(doit,target,add_mode))
		b.bind(on_failed=doerr)
		b.widgetBuild(d)
			
	def getActionData(self, widget:Widget, desc, *args):
		data = {}
		rtdesc = self.build_rtdesc(desc)
		if rtdesc:
			rt = self.get_rtdata(widget, rtdesc, *args)
			if rt:
				data.update(rt)
			if desc.get('keymapping'):
				data = keyMapping(data, desc.get('keymapping'))
		Logger.info('getActionData():rtdesc=%s, data=%s', rtdesc, data)
		return data

	def registedfunctionAction(self, widget:Widget, desc, *args):
		target = self.get_target(widget, desc)
		rf = RegisterFunction()
		name = desc.get('rfname')
		func = rf.get(name)
		if func is None:
			Logger.info('Block: desc=%s rfname(%s) not found', 
						str(desc), name)
			raise Exception('rfname(%s) not found' % name)

		params = desc.get('params',{}).copy()
		d = self.getActionData(widget,desc, *args)
		if d:
			params.update(d)
		func(target, *args, **params)

	def scriptAction(self, widget:Widget, desc, *args):
		script = desc.get('script')
		if not script:
			Logger.info('Block: scriptAction():desc(%s) target not found',
							str(desc))
			return 
		target = self.get_target(widget, desc)
		d = self.getActionData(widget,desc, *args)
		ns = {
			"self":target,
			"args":args,
			"kwargs":d
		}
		if d:
			ns.update(d)
		try:
			self.eval(script, ns)
		except Exception as e:
			print_exc()
			print(e)

	def build_rtdesc(self, desc):
		rtdesc = desc.get('rtdata')
		if not rtdesc:
			if desc.get('datawidget'):
				rtdesc = {}
				rtdesc['target'] = desc['datawidget']
				if desc.get('datascript'):
					rtdesc['script'] = desc.get('datacript')
				else:
					rtdesc['method'] = desc.get('datamethod', 'getValue')
				rtdesc['kwargs'] = desc.get('datakwargs', {})
		return rtdesc

	def get_rtdata(self, widget:Widget, desc, *args):
		"""
		desc descript follow attributes
		{
			"target", a target name from widget
			"method", a method name, default is 'getValue',
			"script", script to handle the data
			"kwargs":" a dict arguments, default is {}
		}
		"""
		
		if desc is None or desc == {}:
			print('get_rtdata():desc is None')
			return {}
		kwargs = desc.get('kwargs', {})
		if not isinstance(kwargs, dict):
			kwargs = {}
		w = None
		if len(args) > 0:
			w = args[0]
		target = desc.get('target')
		if target:
			# w = Blocks.getWidgetById(target, from_widget=widget)
			w = Blocks.findWidget(target, from_widget=widget)
		if w is None:
			print('get_rtdata():w is None', desc)
			return {}
		script = desc.get('script')
		if script:
			ns = {
				"self":w,
				"args":args,
				"kwargs":kwargs
			}
			d = self.eval(script, ns)
			return d

		method = desc.get('method', 'getValue')
		if not hasattr(w, method):
			print('get_rtdata():method is None', desc, type(w))
			return {}
		f = getattr(w, method)
		try:
			r = f(**kwargs)
			if isinstance(r, dict):
				return r
			if isinstance(r, DictObject):
				return r.to_dict()
			print('get_rtdata():method return is not dict', desc, 'ret=', r, type(r))
			return {}
		except:
			print_exc()
			return {}

	def methodAction(self, widget:Widget, desc, *args):
		method = desc.get('method')
		target = self.get_target(widget, desc)
		if target is None:
			Logger.info('Block: methodAction():desc(%s) target not found',
							str(desc))
			return
		if method is None:
			Logger.info('Block: methodAction():desc(%s) method not found',
							str(desc))
			return
		if hasattr(target,method):
			f = getattr(target, method)
			kwargs = desc.get('options',{}).copy()
			d = self.getActionData(widget,desc, *args)
			if d:
				kwargs.update(d)
			f(*args, **kwargs)
		else:
			alert('%s method not found' % method)

	def widgetBuild(self, desc):
		"""
		desc format:
		{
			widgettype:<registered widget>,
			id:widget id,
			options:{}
			subwidgets:[
			]
			binds:[
			]
		}
		"""
		def doit(desc):
			if isinstance(desc,DictObject):
				desc = desc.to_dict()
			if not isinstance(desc,dict):
				Logger.info('Block: desc(%s) must be a dict object(%s)',
							desc,type(desc))
				return None

			# desc = self.valueExpr(desc)
			try:
				widget = self.w_build(desc)
				self.dispatch('on_built',widget)
				if hasattr(widget,'ready'):
					widget.ready = True
				return widget
			except Exception as e:
				print_exc()
				self.dispatch('on_failed',e)
				return None

		if isinstance(desc,DictObject):
			desc = desc.to_dict()
		if not isinstance(desc, dict):
			print('widgetBuild1: desc must be a dict object',
						desc,type(desc))
			self.dispatch('on_failed',Exception('miss url'))
			return

		widgettype = desc.get('widgettype')
		while widgettype == "urlwidget":
			opts = desc.get('options',{}).copy()
			extend = desc.get('extend')
			addon = None
			if desc.get('extend'):
				addon = desc.get('extend').copy()
			url = opts.get('url')
			if url is None:
				self.dispatch('on_failed',Exception('miss url'))
				return
			
			if opts.get('url'):
				del opts['url']
			rtdesc = self.build_rtdesc(opts)
			if rtdesc:
				rtdata = self.get_rtdata(None, rtdesc)
				params = opts.get('params', {})
				params.update(rtdata)
				opts['params'] = params

			desc = self.getUrlData(url,**opts)
			if not (isinstance(desc, DictObject) or \
							isinstance(desc, dict)):
				print('Block2: desc must be a dict object',
								desc,type(desc))
				self.dispatch('on_failed',Exception('miss url'))
				return

			if addon:
				desc = dictExtend(desc,addon)
			widgettype = desc.get('widgettype')
		if widgettype is None:
			print('Block3: desc must be a dict object not None')
			return None
		return doit(desc)
	
	@classmethod
	def findWidget(self, id:str, from_widget:Widget=None) -> Widget:
		"""
		-:find up direct, find widget in ancestor
		+ or empty: find down direct, find widget in descendant
		@:find widget by class
		$ or empty:find widget by id
		.:more than one step.

		"""
		def find_widget_by_id(id, from_widget):
			if id=='self':
				return from_widget
			if hasattr(from_widget,'widget_id'):
				if from_widget.widget_id == id:
					return from_widget
			if hasattr(from_widget, id):
				w = getattr(from_widget,id)
				return w
			return None

		def find_widget_by_class(klass, from_widget):
			if from_widget.__class__.__name__ == klass:
				return from_widget
			for k, v in from_widget.__dict__.items():
				if isinstance(v, Widget):
					if v.__class__.__name__ == klass:
						return v
			return None
		
		def _find_widget(name, from_widget, dir='down', 
					find_func=find_widget_by_id):
			w = find_func(name, from_widget)
			if w:
				return w
			if dir == 'down':
				children = [i for i in from_widget.children]
				if hasattr(from_widget, 'get_subwidgets'):
					children = from_widget.get_subwidgets()
				Logger.info('children=%s', str(children))
				for c in children:
					ret = _find_widget(name, from_widget=c, dir=dir)
					if ret:
						return ret
			else:
				if isinstance(from_widget, WindowBase):
					return None
				if from_widget.parent:
					return _find_widget(name, 
								from_widget=from_widget.parent,
								dir=dir)
			Logger.info('Block:_find_widget(), %s, %s %s return None',
						name, from_widget.__class__.__name__, dir)
			return None

		def find_widget(step, from_widget):
			dir = 'down'
			if step[0:1] == '-':
				dir = 'up'
				step = step[1:]
			find_func = find_widget_by_id
			if step[0:1] == '@':
				step = step[1:]
				find_func = find_widget_by_class
			return _find_widget(step, from_widget=from_widget, 
						dir=dir, find_func=find_func)

		steps = id.split('.')
		w = from_widget
		for i, s in enumerate(steps):
			if i==0:
				app = App.get_running_app()
				fid = s
				if fid == '/self' or fid == 'root':
					w = app.root
					if len(steps) == 1:
						return from_widget
					continue
				if fid == 'Window':
					w = Window
					if len(steps) == 1:
						return w
					continue
				if fid == 'app':
					return app

			w = find_widget(s, w)
			if not w:
				return None
		return w
			
	@classmethod
	def getWidgetById(self, id:str, from_widget:Widget=None) -> Widget:
		return Blocks.findWidget(id, from_widget=from_widget)

	def on_built(self,v=None):
		return

	def on_failed(self,e=None):
		return

	def buildKlass(self, desc):
		"""
		desc = {
			"classname":"MyClass",
			"base":["Box", "WidgetReady"],
			"properties":[
				{
					"name":"aaa",
					"type":"str",
					"default":1
				}
			]
			"subwidgets":[
			],
			
		"""
		codes = """
{% for b in bases %}
{{b}} = Factory.{{b}}
{% endfor %}

class {{ classname }}({% for b in bases -%}{{b}}{% endfor %})
{% for p in properties %}
	{{p.name}} {{p.type}}({{p.default}})
{% endfor %}
	def __init__(self, **kw):
		super({{classname}}, self).__init__(**kw)

"""

Factory.register('Blocks',Blocks)
Factory.register('Video',Video)
Factory.register('OrientationLayout', OrientationLayout)



from kivy.utils import platform
from .responsivelayout import VResponsiveLayout
from .utils import isHandHold
from .paging import RelatedLoader

"""
options format
{
	box_width:optional, if set, the child width
	dataurl: url to get data
	page_rows:records per read 
	viewer:{
		widgettype:an class name to show the object
		options:{
		}
	},
	binds:[
		bind info
	]
}
	
"""

class ObjectViewer(VResponsiveLayout):
	def __init__(self, dataurl=None, viewer={}, 
					page_rows=25,params={},**options):
		super().__init__(**options)
		self.options = options
		self.dataUrl = dataurl
		self.params = params
		self.viewer = viewer
		self.page_rows = page_rows
		self.initflag = False
	

	def setScrollPos(self,pos):
		print('setScrollPos(),pos=',pos)
		self.scroll_y = pos

	def sizeChangedWork(self,v=None):
		super().sizeChangedWork()
		if not self.initflag:
			self.initflag = True
			self.loader = RelatedLoader(
							adder=self.showObject,
							remover=self.remove_widget,
							locater=self.setScrollPos,
							page_rows=self.page_rows,
							dataurl=self.dataUrl,
							target=self,
							params=self.params,
							method=self.options.get('method','GET')
			)
			self.loader.loadPage(1)
			print('ObjectViewer:::::::::::::self.size=',self.size)
	
	def on_scroll_stop(self,o,v=None):
		return 
		print('on_scroll_stop(), self.scroll_y=', self.scroll_y,
				self._inner.pos,
				self._inner.size,
				self.size)
		if self.scroll_y >=0.999:
			print('on_scroll_stop(), move_up')
			self.loader.loadPreviousPage()
		if self.scroll_y <= 0.001:
			print('on_scroll_stop(), move_down')
			self.loader.loadNextPage()

	def showObject(self,rec,**kw):
		pass



from pyler import gps

class GPS:
	def __init__(self):
		gps.configure(on_location=self.on_location, on_status=self.on_status)
		gps.start()

	def on_location(self, **kw):
		pass
	
	def on_status(self,**kw):
		pass
	def __del__(self):
		gps.stop()





from functools import partial
import json
from pythonosc import dispatcher, osc_server, udp_client
from appPublic.sockPackage import get_free_local_addr
from appPublic.background import Background
from kivy.event import EventDispatcher


class PlayerOSCServer(EventDispatcher):
	def __init__(self,playerid, cmds=[]):
		EventDispatcher.__init__(self)
		self.playerid = playerid
		dispatch = dispatcher.Dispatcher()
		self.ip,self.port = get_free_local_addr()
		self.server = osc_server.BlockingOSCUDPServer( (self.ip, self.port), dispatch)
		self.osc_dispatch = dispatch
		self.commands = cmds
		for cmd in self.commands:
			self.map(cmd)
		
	def on_osc_event(self, cmd, *args):
		print('PlayerOSCServer():on_osc_event():cmd=', cmd, 'args=',args)

	def action_event(self,cmd,*args):
		self.dispatch('on_%s' % cmd, cmd, *args)

	def start(self):
		self.thread = Background(self.server.serve_forever)
		self.thread.daemon_threads = True
		self.thread.start()

	def info(self):
		return {
			"playerid":self.playerid,
			"ip":self.ip,
			"port":self.port,
			"commands": self.commands
		}

	def map(self,cmd):
		event_name = 'on_%s' % cmd
		on_f = partial(self.on_osc_event, cmd)
		setattr(self,event_name, on_f)
		self.register_event_type(event_name)
		self.osc_dispatch.map( '/%s' % cmd,partial(self.action_event, cmd))
			
	def stop(self):
		self.server.shutdown()
		self.server.server_close()
		self.thread.join(5)

	def send_message(self, api, data, ip, port):
		client = udp_client.SimpleUDPClient(ip, port)  # Create client
		t = json.dumps(data)
		client.send_message('/%s' % api, t)
		

class PlayerOSCClient:
	def __init__(self, ip,port):
		self.client = udp_client.SimpleUDPClient(ip, port)  # Create client

	def play(self, mrec):
		t = json.dumps(mrec)
		self.client.send_message("/play", t)

	def mute(self):
		self.client.send_message('/mute')

	def menu(self):
		self.client.send_message('/menu')

	def pause(self):
		self.client.send_message('/pause')

	def up(self):
		self.client.send_message('/up')





import time
import numpy as np
from ffpyplayer.player import MediaPlayer
from ffpyplayer.tools import set_log_callback

from kivy.factory import Factory
from kivy.app import App
from kivy.uix.image import Image
from kivy.uix.widget import Widget
from kivy.clock import Clock
from kivy.properties import StringProperty, BooleanProperty, \
			OptionProperty, NumericProperty
from kivy.graphics.texture import Texture
from kivy.graphics import Color, Line
from kivyblocks.ready import WidgetReady


class FFVideo(WidgetReady, Image):
	v_src = StringProperty(None)
	status = OptionProperty('stop', \
			options=['stop', 'play', 'pause'])
	play_mode = OptionProperty('normal', \
			options=['normal', 'preview', 'audioonly'])
	header_callback = StringProperty(None)
	audio_id = NumericProperty(None)
	duration = NumericProperty(-1)
	position = NumericProperty(-1)
	volume = NumericProperty(-1)
	framerate = NumericProperty(180)
	in_center_focus = BooleanProperty(False)

	def __init__(self, **kwargs):
		self._player = None
		self._volume = 0
		self._position = 0
		self._update_task = None
		self._texture = None
		self.is_black = False
		self.videosize = None
		self.timeperiod = 1.0 / self.framerate
		self.ff_opts = {}
		self.lib_opts = {}
		self.headers_pattern = {}
		super(FFVideo, self).__init__(**kwargs)
		self.set_black()
		self.register_event_type('on_frame')
		self.register_event_type('on_open_failed')
		self.register_event_type('on_leave_focus')
		self.register_event_type('on_enter_focus')
		self.register_event_type('on_load_success')
		self.register_event_type('on_startplay')

	def on_open_failed(self, *args):
		pass

	def on_load_success(self, *args):
		pass

	def on_enter_focus(self, *args):
		pass

	def on_leave_focus(self, *args):
		pass

	def check_focus(self,*args):
		if not self.parent:
			self.in_center_screen = False
			return
		if not self.get_root_window():
			self.in_center_screen = False

		w = self.parent
		pos = w.to_widget(*Window.center)
		if self.collide_point(*pos):
			self.in_center_screen = True
		else:
			self.in_center_screen = False

	def set_ff_opt(self, k,v):
		self.ff_opts.update({k:v})

	def set_lib_opt(self, k, v):
		self.lib_opts.update({k:v})

	def set_pattern_header(self, pattern, k,v):
		dic = self.headers_pattern.get(pattern,{})
		dic.update({k:v})
		self.headers_pattern[pattern] = dic

	def _get_spec_headers(self, filename):
		for p in self.headers_pattern.keys():
			if filename.startswith(p):
				headers = self.headers_pattern[p]
				headers_str = ''.join([f'{k}:{v}\r\n' for k,v in headers.items()])
				return headers_str
		return None

	def on_volume(self, *args):
		if self._player is None:
			return

		self._player.set_volume(self.volume)
		self._volume = self.volume
		
	def on_status(self, *args):
		print('on_status called, ', self.status)
		if self._player is None:
			return

		if self.status == 'play':
			self._player.set_pause(False)
		elif self.status == 'pause':
			self._player.set_pause(True)
		elif self.status == 'stop':
			self._play_stop()
		else:
			pass

	def on_parent(self, *args):
		if self.parent:
			return
		self._play_stop()
		del self

	def on_startplay(self, *args):
		pass

	def on_frame(self, *args):
		if self._player is None:
			return
		if not self.playing:
			self.dispatch('on_startplay')
			self._player.request_channel( \
								'audio', 'open', self.audio_id)
			self.seek(self.position)
			self.playing = True
		if self.duration > 0:
			p = self._position / self.duration * self.width
			self.canvas.after.clear()
			with self.canvas.after:
				Color(1,1,1,1)
				Line()
				Line(points=[0, 0, self.width, 0], width=1)
				Color(1,0,0,1)
				Line(points=[0,1,p,1], width=1)
		self.position = self._position

	def __del__(self):
		if self._update_task:
			self._update_task.cancel()
			self._update_task = None
		if self._player:
			self._player.close_player()
			self._player = None

	def set_volume(self, v):
		if self.play_mode == 'preview':
			return
		if self._player is None:
			return
		if self.status != 'play':
			return
		if v > 1.0:
			v = 1.0
		if v < 0.0:
			v = 0.0
		self.volume = v

	def seek(self, pts):
		if self.play_mode == 'preview':
			return
		if self._player is None:
			return
		if self.status != 'play':
			return
		self._player.seek(pts, relative=False)
		self.last_frame, self.timepass = self._player.get_frame()
		self._position = self._player.get_pts()

	def mute(self, flag=None):
		if self.play_mode == 'preview':
			return
		if self._player is None:
			return
		if self.status != 'play':
			return
		x = self._player.get_mute()
		print('Video(), mute=', x)
		self._player.set_mute(not x)
		
	def switch_audio(self):
		if self.play_mode == 'preview':
			return
		if self._player is None:
			return
		if self.status != 'play':
			return
		self._player.request_channel('audio', action='cycle')

	def on_v_src(self, o, src):
		self._play_stop()
		self.playing = False

		ff_opts = {
			'pause':False
		}

		if self.play_mode == 'preview':
			ff_opts['lowres'] = 2 # 1/4 size
			ff_opts['an'] = True
		elif self.play_mode == 'audioonly':
			ff_opts['vn'] = True
		ff_opts.update(self.ff_opts)

		lib_opts = {
		}
		lib_opts.update(self.lib_opts)
		heads = self._get_spec_headers(self.v_src)
		if heads:
			lib_opts.update({'headers':heads})
			ff_opts.update({'headers':heads})

		print('ff_opts=', ff_opts)
		print('lib_opts=', lib_opts)
		# self._player = MediaPlayer(self.v_src) 
		self._player = MediaPlayer(self.v_src, ff_opts=ff_opts, \
						lib_opts=lib_opts) 
		self._play_start()

	def file_opened(self, files):
		self.v_src = files[0]

	def play(self):
		if self._player is None:
			return
		# self._player.set_pause(False)
		self.status = 'play'

	def pause(self):
		if self._player is None:
			return
		# self._player.set_pause(True)
		self.status = 'pause'

	def _play_start(self):
		self.timepass = 0.0
		self.last_frame = None
		self.is_black = False
		self._update_task = Clock.schedule_interval(self._update, self.timeperiod)

	def _get_video_info(self):
		if not self.playing:
			meta = self._player.get_metadata()
			self.duration = meta['duration']
			self._out_fmt = meta['src_pix_fmt']
			self.frame_rate = meta['frame_rate']
			self.videosize = meta['src_vid_size']

	def _play_stop(self):
		if self._player is None:
			return
		self._update_task.cancel()
		self._update_task = None
		self._player.close_player()
		self._player = None
		self.timepass = 0
		self.next_frame = None
		self.duration = -1
		self.position = 0
		self.frame_rate = None
		self.videosize = None
		
	def on_size(self, *args):
		if self._player is None:
			return
		if self.videosize == None:
			return
		w, h = self.videosize
		r = w / h
		r1 = self.width / self.height
		if r1 >= r:
			self._player.set_size(-1, self.height)
		else:
			self._player.set_size(self.width, -1)

	def set_black(self):
		if self.is_black:
			return
		image_texture = Texture.create(
            size=self.size, colorfmt='rgb')
		buf = b'\x00' * int(self.width * self.height * 3)
		image_texture.blit_buffer(buf, colorfmt='rgb', bufferfmt='ubyte')
		self.texture = image_texture
		self.is_black = True

	def show_yuv420(self, img):
		w, h = img.get_size()
		w2 = int(w / 2)
		h2 = int(h / 2)
		self._tex_y = Texture.create(
			size=(w, h), colorfmt='luminance')
		self._tex_u = Texture.create(
			size=(w2, h2), colorfmt='luminance')
		self._tex_v = Texture.create(
			size=(w2, h2), colorfmt='luminance')
		self._fbo = fbo = Fbo(size=(w, h))
		with fbo:
			BindTexture(texture=self._tex_u, index=1)
			BindTexture(texture=self._tex_v, index=2)
			Rectangle(size=fbo.size, texture=self._tex_y)
		fbo.shader.fs = VideoFFPy.YUV_RGB_FS
		fbo['tex_y'] = 0
		fbo['tex_u'] = 1
		fbo['tex_v'] = 2
		self._texture = fbo.texture
		dy, du, dv, _ = img.to_memoryview()
		if dy and du and dv:
			self._tex_y.blit_buffer(dy, colorfmt='luminance')
			self._tex_u.blit_buffer(du, colorfmt='luminance')
			self._tex_v.blit_buffer(dv, colorfmt='luminance')
			self._fbo.ask_update()
			self._fbo.draw()
		texture.flip_vertical()
		self.texture = texture

	def show_others(self, img):
		w, h = img.get_size()
		texture = Texture.create(size=(w, h), colorfmt='rgb')
		texture.blit_buffer(
				img.to_memoryview()[0], colorfmt='rgb',
				bufferfmt='ubyte')
		texture.flip_vertical()
		self.texture = texture
		# print('img_size=', w, h, 'window size=', self.size)

	def _update(self, dt):
		if self.last_frame is None:
			frame, val = self._player.get_frame()
			if val == 'eof':
				self.status = 'stop'
				self.set_black()
				return
			if val == 'pause':
				self.status = 'pause'
				return
			if frame is None:
				self.set_black()
				return
			self.last_frame = frame
			self.video_ts = val
			self._get_video_info()
			
		self.timepass += self.timeperiod
		self._position = self._player.get_pts()
		self._volume = self._player.get_volume()
		if self.timepass < self.video_ts:
			return

		if self.timepass > self.video_ts +0.2:
			self.last_frame = None
			return

		self.status = 'play'
		img, t = self.last_frame
		if self._out_fmt == 'yuv420p':
			self.show_yuv420(img)
		else:
			self.show_others(img)
		self.dispatch('on_frame', self.last_frame)
		self.last_frame = None




# -*- coding: utf-8 -*-

try:
	from setuptools import setup
except ImportError:
	from distutils.core import setup
# from Cython.Build import cythonize

with open('kivyblocks/version.py', 'r') as f:
	x = f.read()
	y = x[x.index("'")+1:]
	z = y[:y.index("'")]
	version = z
with open("README.md", "r") as fh:
    long_description = fh.read()


description = "kivy blocks is a tool to build kivy ui with json format uidesc files"
author = "yumoqing"
email = "yumoqing@icloud.com"

package_data = {
	"kivyblocks":[
		'imgs/*.png', 
		'imgs/*.atlas', 
		'imgs/*.gif',
		'imgs/*.jpg',
		'ttf/*.ttf', 
		'ui/*.uidesc',
		'xcamera/xcamera.kv',
		# 'image_processing/cascades/haarcascade_frontalface_default.xml',
		'xcamera/data/*'
	],
}

setup(
    name="kivyblocks",
	# ext_modules= cythonize( [ ]),
	ext_modules= [],
    version=version,
    # uncomment the following lines if you fill them out in release.py
    description=description,
	long_description=long_description,
	long_description_content_type="text/markdown",
    author=author,
    author_email=email,
   
    install_requires=[
	"kivy",
	"appPublic",
	"sqlor"
    ],
    packages=[
		'kivyblocks',
		# 'kivyblocks.image_processing',
		'kivyblocks.mapview',
		'kivyblocks.uitype',
		'kivyblocks.widgetExt'
		# 'kivyblocks.xcamera'
	],
    package_data=package_data,
    keywords = [
    ],
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
	platforms= 'any'
)



from math import floor
from kivy.factory import Factory
from kivy.properties import NumericProperty, StringProperty, \
		DictProperty, ListProperty, OptionProperty, BooleanProperty
from kivy.logger import Logger
from kivy.uix.textinput import TextInput
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.stacklayout import StackLayout
from kivy.core.window import Window
from kivy.graphics import Color
from .responsivelayout import VResponsiveLayout
from .baseWidget import *
from .utils import *
from .i18n import I18n
from .toolbar import Toolbar
from .color_definitions import getColors
from .dataloader import DataGraber
from .ready import WidgetReady
from .widget_css import WidgetCSS
from .dateinput import DateInput
from .uitype.factory import UiFactory      # import get_input_builder

"""
form options
{
	"toolbar":
	"dataloader":{
		"dataurl":"first"
		"datatargment":"second",
		"datarfname":"third"
		"params":
		"method"
	}
	"cols":"1"
	"labelwidth":
	"textsize":
	"inputheight":
	"inline":True,False
	"fields":[
		{
			"name":
			"label":
			"datatype",
			"uitype",
			"uiparams",
			"default",
			"readonly",
			"required"
		},
	]
	"binds":[
	]
}
"""

class InputBox(BoxLayout):
	def __init__(self, form, **options):
		self.input_widget = None
		self.form = form
		self.options = options
		self.uitype = options.get('uitype',options.get('datatype','str'))
		self.size_hint = (None, None)
		height = self.form.inputheight
		if self.uitype == 'text':
			if not options.get('height'):
				self.options['height'] = 4
			height = self.options.get('height',4)

		self.labelwidth = self.form.labelwidth
		kwargs = {
			"orientation":'horizontal',
		}
		if height<=1:
			kwargs['size_hint_y'] = height
		else:
			kwargs['size_hint_y'] = None
			kwargs['height'] = CSize(height)
		if self.form.input_width <= 1:
			kwargs['size_hint_x'] = self.form.input_width
		else:
			kwargs['size_hint_x'] = None
			kwargs['width'] = self.form.input_width
		super().__init__(**kwargs)
		self.initflag = False
		self.register_event_type("on_datainput")
		self.init()
	
	def focus_act(self, *args):
		pass
		

	def on_size(self, *args):
		if self.input_widget is None:
			return
		self.labeltext.size_hint_x = None
		if self.form.labelwidth < 1:
			self.labeltext.width = self.width * self.form.labelwidth - \
									CSize(1)
		else:
			self.labeltext.width = CSize(self.form.labelwidth)

		self.input_widget.size_hint_x = None
		self.input_widget.width = self.width - self.labeltext.width - CSize(1)

	def on_datainput(self,o,v=None):
		pass
		#print('on_datainput fired ...',o,v)

	def init(self):
		if self.initflag:
			return
		i18n = I18n()
		label = self.options.get('label',self.options.get('name'))
		kwargs = {
			"i18n":True,
			"otext":label,
			"font_size":CSize(1),
		}
		if self.form.labelwidth <= 1:
			kwargs['size_hint_x'] = self.form.labelwidth
		else:
			kwargs['size_hint_x'] = None
			kwargs['width'] = CSize(self.form.labelwidth)

		self.labeltext = Text(**kwargs)
		self.add_widget(self.labeltext)
		if self.options.get('required',False):
			star = Label(text='*',
						color=(1,0,0,1),
						size_hint_x=None,
						width=CSize(1))
			self.add_widget(star)
		
		options = self.options.copy()
		options['hint_text'] = i18n(self.options.get('hint_text'))
		options['size_hint_y'] = None
		options['height'] = self.height
		self.input_widget = UiFactory.build_input_widget(options)
		try:
			self.input_widget.bind(focus=self.focus_act)
		except:
			pass
		self.input_widget.height = self.height
		self.input_widget.size_hint_y = None

		if self.form.labelwidth < 1:
			self.input_widget.size_hint_x = 1 - self.form.labelwidth
		else:
			self.input_widget.size_hint_x = None
			self.input_widget.width = self.width - self.labeltext.width - CSize(1)
		if self.options.get('readonly'):
			self.input_widget.disabled = True
		if self.options.get('value'):
			self.input_widget.setValue(self.options.get('value'))
		elif self.options.get('default_value'):
			self.input_widget.setValue(self.options.get('default_value'))

		self.input_widget.widget_id = self.options['name']
		self.add_widget(self.input_widget)
		if self.labelwidth<=1:
			self.labeltext.size_hint_x = self.labelwidth
			self.input_widget.size_hint_x = 1 - self.labelwidth
		else:
			self.labeltext.size_hint_x = None
			self.labeltext.width = CSize(self.labelwidth)
		self.initflag = True
		self.input_widget.bind(on_focus=self.on_focus)

	def check(self):
		d = self.getValue()
		v = d.get(self.options.get('name'))
		Logger.info('InputWidget() getValue=%s, name=%s',
						v,self.options.get('name'))
		if self.options.get('required',False) and \
				(v == '' or v is None):
			return False

		return True

	def clear(self):
		self.input_widget.setValue('')

	def on_focus(self,o,v):
		if v:
			self.old_value = o.text
		else:
			if self.old_value != o.text:
				self.dispatch('on_datainput',o.text)
		
	def setValue(self,v):
		self.input_widget.setValue(v)
	
	def getValue(self):
		return {self.options.get('name'):self.input_widget.getValue()}
	
	def disable(self,*args,**kwargs):
		self.input_widget.disabled = True

	def enable(self,*args,**kwargs):
		self.input_widget.disabled = False

def defaultToolbar():
	return {
		"img_size_c":1,
		"text_size_c":1,
		"toolbar_orient":"H",
		"tool_orient":"horizontal",
		"tools":[
			{
				"name":"__submit",
				"source_on":blockImage("save.png"),
				"source_off":blockImage("save.png"),
				"label":"submit"
			},
			{
				"name":"__clear",
				"source_on":blockImage("clear.png"),
				"source_off":blockImage("clear.png"),
				"label":"clear"
			}
		]

	}

class Form(WidgetCSS, WidgetReady, BoxLayout):
	cols = NumericProperty(None)
	csscls=StringProperty('formcss')
	input_css=StringProperty('formfieldcss')
	inputwidth=NumericProperty(1)
	inputheight=NumericProperty(3)
	labelwidth=NumericProperty(0.3)
	toolbar_at=OptionProperty('top', \
						options=['top','bottom','left','right'])
	dataloader=DictProperty(None)
	fields = ListProperty([])
	toolbar=DictProperty(None)
	submit=DictProperty(None)
	clear=DictProperty(None)
	notoolbar=BooleanProperty(False)
	def __init__(self, **options):
		self.cols = 1
		self.toolbar_w = None
		if self.toolbar_at in ['top','bottom']:
			options['orientation'] = 'vertical'
		else:
			options['orientation'] = 'horizontal'
		SUPER(Form, self, options)
		self.setup_cols_and_inputwidth()
		self.init()
		self.register_event_type('on_submit')

	def setup_cols_and_inputwidth(self):
		if self.inputwidth > 1:
			self.input_width = CSize(self.inputwidth)
			return
		self.input_width = self.inputwidth

	def set_grid_attrs(self):
		self.fsc.box_width = self.input_width

	def on_size(self, *args):
		if not hasattr(self,'fsc'):
			return

		if not self.notoolbar:
			if self.toolbar_at in ['top', 'bottom']:
				self.fsc.height = self.height - self.toolbar_w.height
				self.fsc.width = self.width
			else:
				self.fsc.width = self.width - self.toolbar_w.width
				self.fsc.height = self.height
		else:
			if self.toolbar_at in ['top', 'bottom']:
				self.fsc.height = self.height
				self.fsc.width = self.width
			else:
				self.fsc.width = self.width
				self.fsc.height = self.height
		self.set_grid_attrs()

	def init(self):
		if not self.notoolbar:
			desc = defaultToolbar()
			desc1 = self.toolbar
			tools = desc['tools'].copy()
			if desc1:
				desc.update(desc1)
				tools = tools + desc1['tools']
			if self.submit:
				kw = self.submit.copy()
				if kw.get('name'):
					del kw['name']
				for t in tools:
					if t['name'] == '__submit':
						t.update(kw)
			if self.clear:
				kw = self.clear.copy()
				if kw.get('name'):
					del kw['name']
				for t in tools:
					if t['name'] == '__clear':
						t.update(kw)
				
			desc['tools'] = tools
			if self.toolbar_at in ['top', 'bottom']:
				desc['toolbar_orient'] = 'H'
			else:
				desc['toolbar_orient'] = 'V'
				desc['tool_orient'] = 'veritcal'
			
			self.toolbar_w = Factory.Toolbar(**desc)
		print('box_width=', self.input_width)
		self.fsc = VResponsiveLayout(
						box_width = self.input_width,
						size_hint=(1,1)
		)

		if self.toolbar_at in ['top', 'left'] and self.toolbar_w:
			self.add_widget(self.toolbar_w)
		self.add_widget(self.fsc)
		if self.toolbar_at in ['bottom', 'right'] and self.toolbar_w:
			self.add_widget(self.toolbar_w)
		
		self.fieldWidgets=[]
		for f in self.fields:
			w = InputBox(self, **f)
			self.fsc.add_widget(w)
			self.fieldWidgets.append(w)
			w.bind(on_ready=self.makeInputLink)
		wid = Factory.Blocks.getWidgetById('__submit',from_widget=self)
		if wid:
			wid.bind(on_press=self.on_submit_button)
		wid = Factory.Blocks.getWidgetById('__clear',from_widget=self)
		if wid:
			wid.bind(on_press=self.on_clear_button)
		if self.dataloader:
			self.loader = DataGraber(**self.dataloader)
			d = self.loader.load()
			if d:
				self.setValue(d)
		self.on_size()
			
	def makeInputLink(self,o,v=None):
		self.readiedInput += 1
		if self.readiedInput >= len(self.fields):
			p = self.fieldWidgets[0]
			for w in self.fieldWidgets[1:]:
				p.input_widget.focus_next = w.input_widget
				w.input_widget.focus_previous = p.input_widget
				p = w

	def setValue(self,d):
		for f in self.fieldWidgets:
			v = f.getValue()
			for k in v.keys():
				f.setValue({k:d[k]})

	def getValue(self):
		d = {}
		for f in self.fieldWidgets:
			v = f.getValue()
			d.update(v)
		return d

	def checkData(self):
		for w in self.fieldWidgets:
			if not w.check():
				w.input_widget.focus = True
				Logger.info('kivyblcks: input check false')
				return False
		Logger.info('kivyblcks: input check success')
		return True

	def on_submit(self,o, v=None):
		return False

	def on_submit_button(self,o,v=None):
		Logger.info('kivyblcks: submit button press')
		if not self.checkData():
			Logger.info('kivyblocks: CheckData False')
			return
		d = self.getValue()
		Logger.info('kivyblocks: fire on_submit')
		self.dispatch('on_submit',d)

	def on_clear_button(self,o,v=None):
		for fw in self.fieldWidgets:
			fw.clear()

class StrSearchForm(BoxLayout):
	def __init__(self,img_url=None,**options):
		self.name = options.get('name','search_string')
		BoxLayout.__init__(self,orientation='horizontal',size_hint_y=None,height=CSize(3))
		self.input_css = options.get('input_css', 'input')
		i18n = I18n()
		self.input_widget = StrInput(
				text='',
				multiline=False,
				csscls=self.input_css,
				hint_text=i18n(options.get('tip',options.get('hint_text',''))),
				allow_copy=True,
				font_size=CSize(1),
				size_hint_y=None,
				size_hint_x=1,
				height=3)
		self.add_widget(self.input_widget)
		self.register_event_type('on_submit')
		v = options.get('value',options.get('default_value',''))
		self.input_widget.setValue(v)
		self.input_widget.bind(on_text_validate=self.submit_input)

	def getValue(self):
		d = {
			self.name:self.input_widget.text
		}
		return d

	def setValue(self, d):
		if isinstance(d,str):
			self.input_widget.text = d
		if isinstance(d,{}):
			self.input_widget.text = d.get(self.name,'')

	def submit_input(self,o,v=None):
		text = self.input_widget.text
		if text != '':
			d = {
				self.name:text
			}
			self.dispatch('on_submit',d)

	def on_submit(self,v=None):
		pass



try:
	import ujson as json
except:
	import json

import time
import zlib
from kivy.logger import Logger
from kivy.event import EventDispatcher
from kivy.clock import Clock
from appPublic.udp_comm import UdpComm
from appPublic.dataencoder import DataEncoder


class UdpWidget(EventDispatcher):
	def __init__(self, udp_port=55000, cert_file=None, commands=[],
					**kw):
		super(UdpWidget, self).__init__(**kw)
		self.udp_port = udp_port
		self.commands = commands
		self.block_commands = []
		self.udp_transport = UdpComm(udp_port, self.comm_callback)
		host = self.udp_transport.host
		self.dataencoder = DataEncoder(host, self.get_peer_pubkey)
		self.inner_handlers = {
			'get_pubkey':self.resp_pubkey,
			'set_pubkey':self.set_pubkey
		}
		for cmd in self.commands:
			evt_name = 'on_%s' % cmd
			setattr(self, evt_name, self.my_event_handler)
			self.register_event_type(evt_name)
			# print('udp_widget.py:register', evt_name, self.my_event_handler)

		self.get_peer_pubkey()
		Clock.schedule_once(self.get_peer_pubkey_loop, 2)

	def block_command(self, cmd):
		if cmd not in self.command:
			return
		if cmd in self.block_command:
			return
		self.block_commands.append(cmd)

	def unblock_command(self, cmd):
		if cmd not in self.block_command:
			return
		self.block_commands = [ c for c in self.block_command if c!=cmd ]

	def get_peer_pubkey_loop(self, t):
		self.get_peer_pubkey()

	def get_peer_pubkey(self, peer_id=None, timeout=1):
		# print('get_peer_pubkey(), called..')
		d = {
			'c':'get_pubkey',
			'd':{
				'pubkey':self.dataencoder.my_text_publickey()
			}
		}
		bd = self.dataencoder.pack('none', d, uncrypt=True)
		self.udp_transport.broadcast(bd)
		if peer_id is None:
			# print('get_peer_pubkey():return')
			return
		t = t1 = time.time()
		t1 += timeout
		while t1 > t:
			time.sleep(0.1)
			t = time.time()
			if self.dataencoder.exist_peer_publickeys(peer_id):
				return self.dataencder.public_keys[peer_id]
		raise Exception('timeout')

	def comm_callback(self, data, addr):
		# print('comm_callback():', data, 'addr=', addr)
		d = self.dataencoder.unpack(addr[0], data)
		if d is None:
			# print('comm_callback(): d is None')
			return
		if not isinstance(d, dict):
			# print('comm_callback(): d is not a dict data')
			return
		# print('received: data=', d)
		cmd = d['c']
		f = self.inner_handlers.get(cmd)
		if f:
			f(d, addr)
			# print('comm_callback():inner callback called(),', cmd)
			return
		if cmd in self.block_commands:
			# print('comm_callback():', cmd, 'is blocked')
			return
		evt_name = 'on_%s' % cmd
		evt_data = {
			'd': d,
			'addr': addr
		}
		# print('udp_widget.py dispatch', evt_name, evt_data)
		self.dispatch(evt_name, evt_data)

	def my_event_handler(self, *args):
		pass

	def set_pubkey(self, data, addr):
		pk = data['d']['pubkey']
		id = addr[0]
		# print('set_pubkey(): ', id, pk)
		self.dataencoder.set_peer_text_pubkey(id, pk)

	def resp_pubkey(self, data, addr):
		self.set_pubkey(data, addr)
		# print('resp_pubkey():', addr[0])
		data = {
			'c':'set_pubkey',
			'd':{
				'pubkey':self.dataencoder.my_text_publickey()
			}
		}
		self.send(addr[0], data, uncrypt=True)

	def broadcast(self, data):
		for peer in self.dataencoder.public_keys.keys():
			self.send(peer, data)

	def send(self, peer_id, data, uncrypt=False):
		# print('send():', peer_id, data)
		d = self.dataencoder.pack(peer_id, data, uncrypt=uncrypt)
		addr = (peer_id, self.udp_port)
		self.udp_transport.send(d, addr)

	def stop(self):
		self.udp_transport.stop()



import sys
import math
from traceback import print_exc

from kivy.properties import ObjectProperty, StringProperty, \
			NumericProperty, BooleanProperty, OptionProperty
from kivy.properties import DictProperty
from kivy.app import App
from kivy.utils import platform
from kivy.uix.button import Button, ButtonBehavior
from kivy.uix.accordion import Accordion,AccordionItem
from kivy.uix.label import Label
from kivy.event import EventDispatcher
from kivy.metrics import dp
from kivy.core.window import Window
from kivy.clock import Clock

from kivy.uix.actionbar import ActionBar,ActionView,ActionPrevious,ActionItem,ActionButton
from kivy.uix.actionbar import ActionToggleButton, ActionCheck,ActionSeparator,ActionGroup

from kivy.uix.anchorlayout import AnchorLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.gridlayout import GridLayout
from kivy.uix.pagelayout import PageLayout
from kivy.uix.scatterlayout import ScatterLayout
from kivy.uix.recycleboxlayout import RecycleBoxLayout
from kivy.uix.relativelayout import RelativeLayout
from kivy.uix.stacklayout import StackLayout

from kivy.uix.checkbox import CheckBox
from kivy.uix.switch import Switch
from kivy.uix.textinput import TextInput
from kivy.uix.dropdown import DropDown
from kivy.uix.tabbedpanel import TabbedPanel,TabbedPanelContent,TabbedPanelHeader,TabbedPanelItem
from kivy.uix.treeview import TreeView
from kivy.uix.image import AsyncImage,Image
from kivy.uix.togglebutton import ToggleButton
from kivy.uix.scrollview import ScrollView
from kivy.uix.splitter import Splitter
from kivy.uix.spinner import Spinner
from kivy.uix.slider import Slider

from kivy.uix.screenmanager import ScreenManager
from kivy.uix.sandbox import Sandbox
from kivy.uix.progressbar import ProgressBar
from kivy.uix.popup import Popup
from kivy.uix.modalview import ModalView
from kivy.uix.filechooser import FileChooser
from kivy.uix.effectwidget import EffectWidget
from kivy.uix.colorpicker import ColorPicker
from kivy.uix.carousel import Carousel
from kivy.uix.bubble import Bubble
from kivy.uix.codeinput import CodeInput
from kivy.graphics import Color, Rectangle
from kivy.properties import ListProperty, DictProperty
from kivy.factory import Factory

from appPublic.dictObject import DictObject
from appPublic.jsonConfig import getConfig
from appPublic.folderUtils import listFile

from .widgetExt.scrollwidget import ScrollWidget
from .widgetExt.binstateimage import BinStateImage
from .widgetExt.jsoncodeinput import JsonCodeInput
from .widgetExt.inputext import FloatInput,IntegerInput, \
		StrInput,SelectInput, BoolInput, Password
from .widgetExt.messager import Messager
from .bgcolorbehavior import BGColorBehavior
from .utils import NeedLogin, InsufficientPrivilege, HTTPError
from .login import LoginForm
from .tab import TabsPanel
from .threadcall import HttpClient
from .i18n import I18n
from .widget_css import WidgetCSS
from .ready import WidgetReady
from .utils import CSize, SUPER
from .swipebehavior import SwipeBehavior

if platform == 'android':
	from .widgetExt.phonebutton import PhoneButton
	from .widgetExt.androidwebview import AWebView

class WrapText(Label):
	def __init__(self, **kw):
		Label.__init__(self, **kw)
		self.bind(width=lambda *x: self.setter('text_size')(self, (self.width, None)),
				texture_size=lambda *x: self.setter('height')(self, self.texture_size[1]))

class StackBox(WidgetCSS, WidgetReady, StackLayout):
	def __init__(self, **kw):
		super().__init__(**kw)

class AnchorBox(WidgetCSS, WidgetReady, AnchorLayout):
	def __init__(self, **kw):
		super().__init__(**kw)

def FloatBox(WidgetCSS, WidgetReady, FloatLayout):
	def __init__(self, **kw):
		super().__init__(**kw)

def RelativeBox(WidgetCSS, WidgetReady, RelativeLayout):
	def __init__(self, **kw):
		super().__init__(**kw)

def GridBox(WidgetCSS, WidgetReady, GridLayout):
	def __init__(self, **kw):
		super().__init__(**kw)

def PageBox(WidgetCSS, WidgetReady, PageLayout):
	def __init__(self, **kw):
		super().__init__(**kw)

def ScatterBox(WidgetCSS, WidgetReady, ScatterLayout):
	def __init__(self, **kw):
		super().__init__(**kw)

class Box(WidgetCSS, WidgetReady, BoxLayout):
	def __init__(self, **kw):
		try:
			SUPER(Box, self, kw)
		except Exception as e:
			print('Box(',kw,') Error')
			raise e

class HBox(Box):
	def __init__(self,**kw):
		kw.update({'orientation':'horizontal'})
		Box.__init__(self, **kw)


class VBox(Box):
	def __init__(self,**kw):
		kw.update({'orientation':'vertical'})
		Box.__init__(self, **kw)

class SwipeBox(SwipeBehavior, Box):
	def __init__(self, **kw):
		SUPER(SwipeBox, self, kw)


class Text(Label):
	lang=StringProperty('')
	otext = StringProperty('')
	def __init__(self,i18n=False, texttype='text', wrap=False,
					fgcolor=None, **kw):
		
		fontsize={'font_size':CSize(1)}
		offset={
			'text':0,
			'title1':CSize(0.6),
			'title2':CSize(0.5),
			'title3':CSize(0.4),
			'title4':CSize(0.3),
			'title5':CSize(0.2),
			'title6':CSize(0.1),
		}
		fontsize = {'font_size': CSize(1) + offset.get(texttype,0)}
		self._i18n = i18n
		self.i18n = I18n()
		self.bgcolor = fgcolor
		kwargs = kw.copy()
		config = getConfig()
		self.wrap = wrap
		if kwargs.get('font_size') and texttype=='text':
			pass
		else:
			kwargs.update(fontsize)
		if not kwargs.get('text'):
			kwargs['text'] = kwargs.get('otext','')
		
		SUPER(Text, self, kwargs)
		if self._i18n:
			self.i18n = I18n()
			self.i18n.addI18nWidget(self)
		if self.wrap:
			self.size_hint_y = None
			self.text_size = self.width, None
			self.bind(width=self.set_widget_height)
		if self.bgcolor:
			self.color = self.bgcolor

	def resize(self, *args):
		if not self.size_hint_y:
			ps = [0,0,0,0]
			if hasattr(self.parent,'padding'):
				ps = self.parent.padding

			self.width = self.parent.width - ps[0] - ps[2]
			if self.width > 0:
				self.set_widget_height()

	def set_widget_height(self, *args):
		if self.width == 0:
			return
		self.text_size = self.width, None
		rows = len(self.text) * (self.font_size * 0.621) / self.width
		rows = math.ceil(rows)
		self.height = rows * self.font_size * 1.5

	def get_wraped_size(self):
		if self.text:
			self._label.refresh()
			return self._label.size
		return (None,None)

	def on_size(self,o,size):
		# super().on_size(o,size)
		if self.wrap:
			font_size = self.font_size
			self.text_size = self.width, None

	def on_otext(self,o,v=None):
		if self._i18n:
			self.text = self.i18n(self.otext)
		else:
			self.text = self.otext
	
	def changeLang(self,lang):
		self.lang = lang

	def on_lang(self,o,lang):
		self.text = self.i18n(self.otext)

class Title1(Text):
	def __init__(self, **kw):
		kw.update({'texttype':'title1','bold':True})
		Text.__init__(self, **kw)

class Title2(Text):
	def __init__(self, **kw):
		kw.update({'texttype':'title2','bold':True})
		Text.__init__(self, **kw)

class Title3(Text):
	def __init__(self, **kw):
		kw.update({'texttype':'title3','bold':True})
		Text.__init__(self, **kw)

class Title4(Text):
	def __init__(self, **kw):
		kw.update({'texttype':'title4','bold':True})
		Text.__init__(self, **kw)

class Title5(Text):
	def __init__(self, **kw):
		kw.update({'texttype':'title5','bold':True})
		Text.__init__(self, **kw)

class Title6(Text):
	def __init__(self, **kw):
		kw.update({'texttype':'title6','bold':True})
		Text.__init__(self, **kw)

class Modal(VBox):
	content = DictProperty(None)
	auto_open = BooleanProperty(True)
	auto_dismiss = BooleanProperty(True)
	position = OptionProperty('tc',options=['tl', 'tc', 'tr',
											'cl', 'cc', 'cr',
											'bl', 'bc', 'br'])

	def __init__(self, **kw):
		SUPER(Modal, self, kw)
		self.register_event_type('on_open')
		self.register_event_type('on_pre_open')
		self.register_event_type('on_pre_dismiss')
		self.register_event_type('on_dismiss')
		if self.content:
			blocks = Factory.Blocks()
			self.content_w = blocks.widgetBuild(self.content)
			if self.content_w:
				self.add_widget(self.content_w)
			else:
				print(content,':cannot build widget')

	def on_touch_down(self, touch):
		if 	not self.collide_point(touch.x, touch.y):
			if self.auto_dismiss:
				self.dispatch('on_pre_dismiss')
				self.dismiss()
				return True
				
		return super().on_touch_down(touch)

	def set_modal_position(self, w):
		xn = self.position[1]
		yn = self.position[0]
		x, y = 0, 0
		if xn == 'c':
			x = (w.width - self.width) / 2
		elif xn == 'r':
			x = w.width - self.width
		if x < 0:
			x = 0
		if yn == 'c':
			y = (w.height - self.height) / 2
		elif yn == 'b':
			y = w.height - self.height
		if y < 0:
			y = 0
		if w == Window:
			self.pos = x, y
		else:
			self.pos = w.pos[0] + x, w.pos[1] + y

	def open(self, widget=None):
		if self.parent:
			return
		self.dispatch('on_pre_open')
		if widget is None:
			widget = Window
		self.set_modal_position(widget)
		Window.add_widget(self)
		self.dispatch('on_open')

	def dismiss(self, *args):
		self.dispatch('on_pre_dismiss')
		self.dispatch('on_dismiss')
		Window.remove_widget(self)
			
	def on_open(self, *args):
		pass

	def on_dismiss(self, *args):
		pass
			
	def on_pre_open(self, *args):
		pass

	def on_pre_dismiss(self, *args):
		pass

	def add_widget(self, w, *args, **kw):
		super().add_widget(w, *args, **kw)
		if self.auto_open:
			self.open()

class TimedModal(Modal):
	show_time = NumericProperty(0)
	def __init__(self, **kw):
		self.time_task = None
		SUPER(TimedModal, self, kw)

	def open(self, *args, **kw):
		if self.time_task is not None:
			self.time_task.cancel()
			self.time_task = None
		if self.show_time > 0:
			self.time_task = \
				Clock.schedule_once(self.dismiss, self.show_time)
		super().open(*args, **kw)

	def dismiss(self, *args, **kw):
		if self.time_task:
			self.time_task.cancel()
			self.time_task = None
		super().dismiss()

class PressableImage(ButtonBehavior,AsyncImage):
	def on_press(self):
		pass

class PressableLabel(ButtonBehavior, Text):
	def on_press(self):
		pass

class FILEDataHandler(EventDispatcher):
	def __init__(self, url, suffixs=[],params={}):
		self.url = url
		self.subfixes=subfixes
		self.params = params
		self.page_rows = self.params.get('page_rows',60)
		self.page = self.params.get('page',1)
		if not url.startswith('file://'):
			raise Exception('%s is not a file url' % url)
		self.files = [i for i in listFile(url[7:],suffixs=suffixs, \
					rescursive=self.params.get('rescursive',False)) ]
		self.total_files = len(self.files)
		x = 0 if self.total_files % self.page_rows == 0 else 1
		self.total_pages = self.total_files / self.page_rows + x
		self.register_event_type('on_success')
		self.register_event_type('on_error')

	def on_success(self,*args):
		return

	def on_error(self, *args):
		return

	def handle(self,params={}):
		d = {}
		d['total'] = self.total_files
		d['rows'] = []
		p = self.params.copy()
		p.update(params)
		page = p.get('page')
		for i in range(self.page_rows):
			try:
				r = self.files[page * self.page_rows + i]
				d['rows'].append({'filename':r,
					'id':r,
					'url':r
				})
			except:
				break
		self.dispatch('on_success',d)

class HTTPDataHandler(EventDispatcher):
	def __init__(self, url='',method='GET',params={},
				headers={},
				files={},
				stream=False
				):
		EventDispatcher.__init__(self)
		self.url = url
		self.method = method
		self.params = params
		self.headers = headers
		self.files=files
		self.stream=stream
		self.register_event_type('on_success')
		self.register_event_type('on_error')

	def on_success(self,*args):
		print('HTTPDataHandler():',self.url,'finished...')
		pass

	def on_error(self,*args):
		pass

	def onSuccess(self,o,v):
		print(self.url,'onSuccess() called')
		self.dispatch('on_success',v)

	def onError(self,o,e):
		print(self.url,'onError():v=',e)
		self.dispatch('on_error',e)
		print_exc()
		print('[****][*********] onError Called',e)

	def redo(self,o,v=None):
		self.handle()

	def handle(self,params={},headers={}):
		p = self.params
		p.update(params)
		h = self.headers
		h.update(headers)
		hc = HttpClient()
		print('HTTPDataHandler():',self.url,'handling....')
		hc(self.url,method=self.method,
						params=p,
						headers=h,
						files=self.files,
						stream=self.stream,
						callback=self.onSuccess,
						errback=self.onError)


class HTTPSeriesData(HTTPDataHandler):
	def __init__(self, rows=20, page=1, **kw):
		super(HTTPSeriesData, self).__init__(**kw)
		self.rows = rows
		self.page = page
		self.maxpage = 999999

	def load(self, page=1, **params):
		self.params.update(params)
		if page == 1:
			self.maxpage = 999999
		self.page = page
		params = {
			'page': self.page,
			'rows': self.rows
		}
		self.handle(params=params)

	def loadPrevious(self, **params):
		self.params.update(params)
		if self.page > 0:
			self.load(page=self.page - 1)

	def loadNext(self, **params):
		self.params.update(params)
		if self.page < self.maxpage:
			self.load(page=self.page + 1)

def getDataHandler(url,**kwargs):
	if url.startswith('file://'):
		return FILEDataHandler(url,**kwargs)
	return HTTPDataHandler(url, **kwargs)

def device_type():
	if platform != "android" and platform != "ios":
		return "desktop"
	elif Window.width >= dp(600) and Window.height >= dp(600):
		return "tablet"
	else:
		return "mobile"

class ExAccordion(WidgetCSS, WidgetReady, Accordion):
	"""
	{
		"widgettype":"Accordion",
		"options":{
			"csscls",
			"items":[{
				"title":"titl1",
				"active":false,
				"widget":{
				}
			}
			]
		}
	}
	"""
	items = ListProperty(None)
	def __init__(self, **kw):
		print('ExAccordion() init...')
		super().__init__(**kw)
		self.width = self.height = 800
		# Clock.schedule_once(self._build_children, 0.1)
		self._build_children()

	def _build_children(self, *args):
		for i in self.items:
			bk = Factory.Blocks()
			w = bk.widgetBuild(i['widget'])
			if w is None:
				continue
			tw = AccordionItem()
			if isinstance(i['title'], str):
				tw.title = i['title']
			else:
				tw.title = bk.widgetBuild(i['title'])
			tw.add_widget(w)
			if i.get('active'):
				self.collapse = False
			else:
				self.collapse = True
			self.add_widget(tw)

class Slider(Carousel):
	"""
	{
		"widgettype":"Slider",
		"options":{
			"direction":"right",
			"loop":true,
			"items":[
				{
					"widgettype":....
				}
			]
		}
	}
	"""
	items = ListProperty(None)
	def __init__(self, **kw):
		print('Carousel pro=', dir(Carousel))
		super().__init__(**kw)
		bk = Factory.Blocks()
		for desc in self.items:
			w = bk.widgetBuild(desc)
			self.add_widget(w)




import android
import android.activity
from os import remove
from jnius import autoclass, cast
from plyer.facades import Camera
from plyer.platforms.android import activity

Intent = autoclass('android.content.Intent')
PythonActivity = autoclass('org.kivy.android.PythonActivity')
MediaStore = autoclass('android.provider.MediaStore')
Uri = autoclass('android.net.Uri')

class AndroidCamera(Camera):
	def _take_picture(self, on_complete, filename=None):
		assert(on_complete is not None)
		self.on_complete = on_complete
		self.filename = filename
		android.activity.unbind(on_activity_result=self._on_activity_result)
		android.activity.bind(on_activity_result=self._on_activity_result)
		intent = Intent(MediaStore.ACTION_IMAGE_CAPTURE)
		uri = Uri.parse('file://' + filename)
		parcelable = cast('android.os.Parcelable', uri)
		intent.putExtra(MediaStore.EXTRA_OUTPUT, parcelable)
		activity.startActivityForResult(intent, 0x123)

	def _take_video(self, on_complete, filename=None):
		assert(on_complete is not None)
		self.on_complete = on_complete
		self.filename = filename
		android.activity.unbind(on_activity_result=self._on_activity_result)
		android.activity.bind(on_activity_result=self._on_activity_result)
		intent = Intent(MediaStore.ACTION_VIDEO_CAPTURE)
		uri = Uri.parse('file://' + filename)
		parcelable = cast('android.os.Parcelable', uri)
		intent.putExtra(MediaStore.EXTRA_OUTPUT, parcelable)
		# 0 = low quality, suitable for MMS messages,
		# 1 = high quality
		intent.putExtra(MediaStore.EXTRA_VIDEO_QUALITY, 1)
		activity.startActivityForResult(intent, 0x123)

	def _on_activity_result(self, requestCode, resultCode, intent):
		if requestCode != 0x123:
			return
		android.activity.unbind(on_activity_result=self._on_activity_result)
		if self.on_complete(self.filename):
			self._remove(self.filename)

	def _remove(self, fn):
		try:
			remove(fn)
		except OSError:
			pass

def instance():
	return AndroidCamera()



from .dataloader import DataLoader
from appPublic.sqlorAPI import runSQLIterator,DBPools,runSQLPaging
from twisted.internet import defer, reactor


class SQLDataLoader(DataLoader):
	def __init__(self,dbdesc,**kw):
		self.dbdesc = dbdesc
		super(DataLoader,self).__init__(**kw)
		
	def _run(self,ns):
		@runSQLIterator
		def _sql(db,ns):
			return self.dbdesc['sqldesc']
		
		d = [ i for i in self._sql(self.dbdesc['db'],ns) ]
		return d
	
	def loadData(self,ns):
		d = defer.maybeDeferred(self._run,ns)
		d.callback(self.dataLaoded)
		return d

class SQLPaggingLoader(SQLDataLoader):
	def _run(self,ns):
		@runSQLIterator
		def _sql(db,ns):
			return self.dbdesc['sqldesc']
		
		d = self._sql(self.dbdesc['db'],ns)
		return d
	


# coding=utf-8

__all__ = ["MapSource"]

import hashlib
from math import atan, ceil, cos, exp, log, pi, tan

from kivy.metrics import dp

from kivyblocks.mapview.constants import (
    CACHE_DIR,
    MAX_LATITUDE,
    MAX_LONGITUDE,
    MIN_LATITUDE,
    MIN_LONGITUDE,
)
from kivyblocks.mapview.downloader import Downloader
from kivyblocks.mapview.utils import clamp


class MapSource:
    """Base class for implementing a map source / provider
    """

    attribution_osm = 'Maps & Data © [i][ref=http://www.osm.org/copyright]OpenStreetMap contributors[/ref][/i]'
    attribution_thunderforest = 'Maps © [i][ref=http://www.thunderforest.com]Thunderforest[/ref][/i], Data © [i][ref=http://www.osm.org/copyright]OpenStreetMap contributors[/ref][/i]'

    # list of available providers
    # cache_key: (is_overlay, minzoom, maxzoom, url, attribution)
    providers = {
        "osm": (
            0,
            0,
            19,
            "http://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png",
            attribution_osm,
        ),
        "osm-hot": (
            0,
            0,
            19,
            "http://{s}.tile.openstreetmap.fr/hot/{z}/{x}/{y}.png",
            "",
        ),
        "osm-de": (
            0,
            0,
            18,
            "http://{s}.tile.openstreetmap.de/tiles/osmde/{z}/{x}/{y}.png",
            "Tiles @ OSM DE",
        ),
        "osm-fr": (
            0,
            0,
            20,
            "http://{s}.tile.openstreetmap.fr/osmfr/{z}/{x}/{y}.png",
            "Tiles @ OSM France",
        ),
        "cyclemap": (
            0,
            0,
            17,
            "http://{s}.tile.opencyclemap.org/cycle/{z}/{x}/{y}.png",
            "Tiles @ Andy Allan",
        ),
        "thunderforest-cycle": (
            0,
            0,
            19,
            "http://{s}.tile.thunderforest.com/cycle/{z}/{x}/{y}.png",
            attribution_thunderforest,
        ),
        "thunderforest-transport": (
            0,
            0,
            19,
            "http://{s}.tile.thunderforest.com/transport/{z}/{x}/{y}.png",
            attribution_thunderforest,
        ),
        "thunderforest-landscape": (
            0,
            0,
            19,
            "http://{s}.tile.thunderforest.com/landscape/{z}/{x}/{y}.png",
            attribution_thunderforest,
        ),
        "thunderforest-outdoors": (
            0,
            0,
            19,
            "http://{s}.tile.thunderforest.com/outdoors/{z}/{x}/{y}.png",
            attribution_thunderforest,
        ),
        # no longer available
        # "mapquest-osm": (0, 0, 19, "http://otile{s}.mqcdn.com/tiles/1.0.0/map/{z}/{x}/{y}.jpeg", "Tiles Courtesy of Mapquest", {"subdomains": "1234", "image_ext": "jpeg"}),
        # "mapquest-aerial": (0, 0, 19, "http://oatile{s}.mqcdn.com/tiles/1.0.0/sat/{z}/{x}/{y}.jpeg", "Tiles Courtesy of Mapquest", {"subdomains": "1234", "image_ext": "jpeg"}),
        # more to add with
        # https://github.com/leaflet-extras/leaflet-providers/blob/master/leaflet-providers.js
        # not working ?
        # "openseamap": (0, 0, 19, "http://tiles.openseamap.org/seamark/{z}/{x}/{y}.png",
        #    "Map data @ OpenSeaMap contributors"),
    }

    def __init__(
        self,
        url="http://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png",
        cache_key=None,
        min_zoom=0,
        max_zoom=19,
        tile_size=256,
        image_ext="png",
        attribution="© OpenStreetMap contributors",
        subdomains="abc",
        **kwargs
    ):
        if cache_key is None:
            # possible cache hit, but very unlikely
            cache_key = hashlib.sha224(url.encode("utf8")).hexdigest()[:10]
        self.url = url
        self.cache_key = cache_key
        self.min_zoom = min_zoom
        self.max_zoom = max_zoom
        self.tile_size = tile_size
        self.image_ext = image_ext
        self.attribution = attribution
        self.subdomains = subdomains
        self.cache_fmt = "{cache_key}_{zoom}_{tile_x}_{tile_y}.{image_ext}"
        self.dp_tile_size = min(dp(self.tile_size), self.tile_size * 2)
        self.default_lat = self.default_lon = self.default_zoom = None
        self.bounds = None
        self.cache_dir = kwargs.get('cache_dir', CACHE_DIR)

    @staticmethod
    def from_provider(key, **kwargs):
        provider = MapSource.providers[key]
        cache_dir = kwargs.get('cache_dir', CACHE_DIR)
        options = {}
        is_overlay, min_zoom, max_zoom, url, attribution = provider[:5]
        if len(provider) > 5:
            options = provider[5]
        return MapSource(
            cache_key=key,
            min_zoom=min_zoom,
            max_zoom=max_zoom,
            url=url,
            cache_dir=cache_dir,
            attribution=attribution,
            **options
        )

    def get_x(self, zoom, lon):
        """Get the x position on the map using this map source's projection
        (0, 0) is located at the top left.
        """
        lon = clamp(lon, MIN_LONGITUDE, MAX_LONGITUDE)
        return ((lon + 180.0) / 360.0 * pow(2.0, zoom)) * self.dp_tile_size

    def get_y(self, zoom, lat):
        """Get the y position on the map using this map source's projection
        (0, 0) is located at the top left.
        """
        lat = clamp(-lat, MIN_LATITUDE, MAX_LATITUDE)
        lat = lat * pi / 180.0
        return (
            (1.0 - log(tan(lat) + 1.0 / cos(lat)) / pi) / 2.0 * pow(2.0, zoom)
        ) * self.dp_tile_size

    def get_lon(self, zoom, x):
        """Get the longitude to the x position in the map source's projection
        """
        dx = x / float(self.dp_tile_size)
        lon = dx / pow(2.0, zoom) * 360.0 - 180.0
        return clamp(lon, MIN_LONGITUDE, MAX_LONGITUDE)

    def get_lat(self, zoom, y):
        """Get the latitude to the y position in the map source's projection
        """
        dy = y / float(self.dp_tile_size)
        n = pi - 2 * pi * dy / pow(2.0, zoom)
        lat = -180.0 / pi * atan(0.5 * (exp(n) - exp(-n)))
        return clamp(lat, MIN_LATITUDE, MAX_LATITUDE)

    def get_row_count(self, zoom):
        """Get the number of tiles in a row at this zoom level
        """
        if zoom == 0:
            return 1
        return 2 << (zoom - 1)

    def get_col_count(self, zoom):
        """Get the number of tiles in a col at this zoom level
        """
        if zoom == 0:
            return 1
        return 2 << (zoom - 1)

    def get_min_zoom(self):
        """Return the minimum zoom of this source
        """
        return self.min_zoom

    def get_max_zoom(self):
        """Return the maximum zoom of this source
        """
        return self.max_zoom

    def fill_tile(self, tile):
        """Add this tile to load within the downloader
        """
        if tile.state == "done":
            return
        Downloader.instance(cache_dir=self.cache_dir).download_tile(tile)




from functools import partial
from ffpyplayer.player import MediaPlayer as OldMediaPlayer
from threading import Thread
from kivy.core.video.video_ffpyplayer import VideoFFPy
from kivy.core.video import video_ffpyplayer

external_ff_opts = {}
external_lib_opts = {}

def set_external_ff_opts(dic):
	external_ff_opts = dic

def set_external_lib_opts(dic):
	external_lib_opts = dic

def MediaPlayer(filename, *args, ff_opts={}, lib_opts={}, **kw):
	ff_opts.update(external_ff_opts)
	lib_opts.update(external_lib_opts)
	if filename.startswith('http://') or \
			filename.startswith('https://'):
		headers_str = get_spec_headers(filename)
		if headers_str is not None:
			lib_opts['headers'] = headers_str

	return OldMediaPlayer(filename, *args, ff_opts=ff_opts,
						lib_opts=lib_opts, **kw)

def hack_mediaplayer():
	video_ffpyplayer.MediaPlayer = MediaPlayer

headers_pattern = {
}

def set_headers_pattern(pattern, headers_str):
	headers_pattern[pattern] = headers_str

def get_spec_headers(filename):
	for p in headers_pattern.keys():
		if p in filename:
			headers = headers_pattern[p]
			headers_str = ';'.join([f'{k}={v}' for k,v in headers.items()])
			return "$'%s'" % headers_str
	return None



from functools import partial
from kivy.logger import Logger
from kivy.uix.behaviors import TouchRippleButtonBehavior
from kivy.graphics import Color, Rectangle
from kivy.uix.boxlayout import BoxLayout
from kivy.factory import Factory
from kivy.uix.image import AsyncImage
from kivy.properties import NumericProperty, DictProperty, \
		ObjectProperty, StringProperty, BooleanProperty

from kivyblocks.ready import WidgetReady
from kivyblocks.bgcolorbehavior import BGColorBehavior
from kivyblocks.utils import CSize, SUPER, blockImage
from kivyblocks.baseWidget import Box, Text
from kivyblocks.widget_css import WidgetCSS
from .uitype.factory import UiFactory, get_value
from .command_action import cmd_action

class TinyText(Text):
	def __init__(self, **kw):
		SUPER(TinyText, self, kw)
		self.size_hint = (None, None)
		self.on_texture_size(self, (1,1))

	def on_texture_size(self, o, ts):
		self.texture_update()
		self.size = self.texture_size

class ClickableBox(TouchRippleButtonBehavior, Box):
	def __init__(self,
				border_width=1,
				**kw):
		SUPER(ClickableBox, self, kw)
		self.size_hint = [None, None]
		self.border_width = border_width
		# self.bind(minimum_height=self.setter('height'))
		# self.bind(minimum_width=self.setter('width'))
		self.bind(children=self.reset_size)

	def reset_size(self, o, s):
		self.size_hint = (None,None)
		if self.orientation == 'horizontal':
			self.height = max([c.height for c in self.children])
			self.width = sum([c.width for c in self.children])
		else:
			self.height = sum([c.height for c in self.children])
			self.width = max([c.width for c in self.children])

	def on_press(self,o=None):
		pass

class _IconBehavior:
	source=StringProperty(None)
	img_kw = DictProperty(None)

	def on_source(self, o, source):
		if self.source is None:
			try:
				self.remove_widget(self.img_w)
			except:
				pass
			self.img_w = None
			return
		self.build_icon()

	def on_img_kw(self, o, img_kw):
		self.build_icon()

	def build_icon(self):
		if not self.source:
			return
		if not self.img_kw:
			self.img_kw = {}

		if self.img_w:
			self.img_w.source = self.source
			for k,v in self.img_kw.items():
				setattr(self.img_w, k,v)
			return
		self.img_w = AsyncImage(source=self.source, **self.img_kw)
		self.add_widget(self.img_w)

class _TextBehavior:
	text = StringProperty(None)
	fontsize = NumericProperty(1)

	def create_text_widget(self):
		if self.text is None:
			return
		self.txt_w = TinyText(otext=self.text, 
						i18n=True,
						font_size=CSize(self.fontsize))
		self.txt_w.font_size = CSize(self.fontsize)
		self.txt_w.bind(texture_size=self.reset_size)
		self.txt_w.size_hint = (None, None)
		self.txt_w.size = self.txt_w.texture_size
		self.add_widget(self.txt_w)

	def on_fontsize(self, o, fs):
		if self.txt_w:
			self.txt_w.font_size = CSize(self.fontsize)
			self.txt_w.texture_update()

	def on_text(self, o, txt):
		if self.text is None:
			return
		if not self.txt_w:
			return
		if self.txt_w:
			self.txt_w.text = self.text
			self.txt_w.texture_update()

class ClickableText(_TextBehavior, ClickableBox):
	def __init__(self, **kw):
		self.txt_w = None
		SUPER(ClickableText, self, kw)
		self.create_text_widget()

class ClickableIconText(_IconBehavior, _TextBehavior, ClickableBox):
	def __init__(self, **kw):
		self.txt_w = None
		self.img_w = None
		SUPER(ClickableIconText, self, kw)
		self.build_icon()
		self.create_text_widget()

class CommandBox(ClickableIconText):
	value = DictProperty(None)
	params = DictProperty(None)
	conform = DictProperty(None)
	target = StringProperty(None)
	datawidget = StringProperty(None)
	datamethod = StringProperty(None)
	url = StringProperty(None)
	rfname = StringProperty(None)
	method = StringProperty(None)
	script = StringProperty(None)
	actions = DictProperty(None)
	def on_press(self, *args):
		if self.datawidget is None:
			self.datawidget = 'self'
			self.datamethod = 'getValue'

		desc = dict(target=self.target,
			datawidget = self.datawidget,
			datamethod = self.method,
			url=self.url,
			params=self.params,
			conform = self.conform,
			rfname = self.rfname,
			method = self.method,
			script = self.script,
			actions = self.actions
		)
		cmd_action(desc, self)

	def setValue(self, value:dict):
		self.value = value

	def getValue(self):
		return self.value

class ToggleText(ClickableText):
	"""
	construct commad
	ToggleText(**{
		"text":"test",
		"css_on":"selected",
		"css_off":"normal"
	})
	"""

	select_state = BooleanProperty(False)
	css_on = StringProperty('default')
	css_off = StringProperty('default')
	def __init__(self, **kw):
		SUPER(ToggleText, self, kw)

	def toggle(self):
		self.select_state = False if self.select_state else True

	def select(self, flag):
		if flag:
			self.select_state = True
		else:
			self.select_state = False

	def state(self):
		return self.select_state

	def on_select_state(self, o, f):
		self.csscls = self.css_on if f else self.css_off
	
class ToggleIconText(ToggleText):
	"""
	{
		"orientation",
		"text",
		"css_on",
		"css_off",
		"source_on",
		"source_off"
	}
	"""
	source_on = StringProperty(None)
	source_off = StringProperty(None)
	source = StringProperty(None)
	img_kw = DictProperty({})
	def __init__(self, **kw):
		self.img_w = None
		SUPER(ToggleIconText, self, kw)
		self.source = self.source_off
		self.img_w = AsyncImage(source=self.source, **self.img_kw)
		self.add_widget(self.img_w, index=len(self.children))
		
	def on_select_state(self, o, f):
		super(ToggleIconText, self).on_select_state(o,f)
		self.source = self.source_on if f else self.source_off

	def on_source(self, o, source):
		if source is None:
			return
		if self.img_w:
			self.img_w.source = self.source

class ClickableImage(_IconBehavior, ClickableBox):
	def __init__(self, **kw):	
		self.img_w = None
		SUPER(ClickableImage, self, kw)
		self.build_icon()

class ToggleImage(ClickableImage):
	source_on = StringProperty(None)
	source_off = StringProperty(None)
	select_state = BooleanProperty(False)
	def __init__(self, **kw):
		SUPER(ToggleImage, self, kw)
		self.source = self.source_on

	def toggle(self):
		self.select_state = False if self.select_state else True

	def select(self, flag):
		if flag:
			self.select_state = True
		else:
			self.select_state = False

	def state(self):
		return self.select_state

	def on_select_state(self, o, f):
		if f:
			self.source = self.source_on
		else:
			self.source = self.source_off

class SingleCheckBox(ClickableBox):
	otext = StringProperty(None)
	select_state = BooleanProperty(False)
	def __init__(self, **kw):	
		self.old_state = None
		self.register_event_type('on_changed')
		SUPER(SingleCheckBox, self, kw)
		self.source_on = blockImage('checkbox-on.png')
		self.source_off = blockImage('checkbox-off.png')
		self.select(False)
		self.create_img_w()
		if self.otext:
			self.txt_w = TinyText(otext=self.otext, i18n=True)
			self.txt_w.bind(size=self.reset_size)
			self.add_widget(self.txt_w)
		self.reset_size(self, None)

	def on_changed(self, o=None):
		pass

	def on_press(self, o=None):
		self.toggle()

	def select(self, f):
		self.select_state = f
	
	def change_img_source(self):
		self.img_w.source = self.source_on if self.select_state else \
							self.source_off

	def toggle(self):
		self.select_state = False if self.select_state else True

	def state(self):
		return self.select_state

	def on_select_state(self, o, f):
		self.change_img_source()
		self.dispatch('on_changed', self.select_state)

	def getValue(self):
		return self.state()

	def setValue(self, v):
		self.select(True if v else False)

	def create_img_w(self):
		if self.select_state:
			self.img_w = AsyncImage(source=self.source_on, 
						size_hint=(None, None),
						size=CSize(1, 1)
						)
		else:
			self.img_w = AsyncImage(source=self.source_off, 
						size_hint=(None, None),
						size=CSize(1, 1)
						)
		self.add_widget(self.img_w)
		self.img_w.bind(size=self.reset_size)

def build_checkbox(desc, rec=None):
	v = get_value(desc, rec)
	if v is None:
		v = False
	x = SingleCheckBox(select_state=v)
	return x

def build_cmdbox_view(desc, rec=None):
	vd = None
	if rec is not None:
		vd = {f:rec.get(f) for f in desc.get('data') }
	kw = {
		'text' : desc.get('label'),
		'source' : desc.get('icon'),
		'img_kw' : desc.get('img_kw',{
			'size_hint':[None, None],
			'height':CSize(1),
			'width':CSize(1)
		}),
		'rfname':desc.get('rfname'),
		'script':desc.get('script'),
		'params':desc.get('params'),
		'conform':desc.get('conform'),
		'target':desc.get('target'),
		'datawidget':desc.get('datawidget'),
		'datamethod':desc.get('datamethod'),
		'method':desc.get('method'),
		'url':desc.get('url')
	}
	if desc.get('uiparams'):
		kw.update(desc.get('uiparams'))
	x = CommandBox(**kw)
	x.setValue(vd)
	return x

UiFactory.register('checkbox', build_checkbox, build_checkbox)
UiFactory.register('cmdbox', build_cmdbox_view, build_cmdbox_view)

r = Factory.register
r('CommandBox', CommandBox)
r('TinyText', TinyText)
r('SingleCheckBox', SingleCheckBox)
r('ClickableBox', ClickableBox)
r('ClickableText',ClickableText)
r('ClickableIconText',ClickableIconText)
r('ToggleText',ToggleText)
r('ToggleIconText',ToggleIconText)
r('ClickableImage',ClickableImage)
r('ToggleImage',ToggleImage)



# coding=utf-8

__all__ = ["Coordinate", "Bbox"]

from collections import namedtuple

Coordinate = namedtuple("Coordinate", ["lat", "lon"])


class Bbox(tuple):
    def collide(self, *args):
        if isinstance(args[0], Coordinate):
            coord = args[0]
            lat = coord.lat
            lon = coord.lon
        else:
            lat, lon = args
        lat1, lon1, lat2, lon2 = self[:]

        if lat1 < lat2:
            in_lat = lat1 <= lat <= lat2
        else:
            in_lat = lat2 <= lat <= lat2
        if lon1 < lon2:
            in_lon = lon1 <= lon <= lon2
        else:
            in_lon = lon2 <= lon <= lon2

        return in_lat and in_lon
